import argparse
import sys
from typing import (
    Iterable, Sequence, Type, TypeVar, Union, Literal, Callable,
    overload, get_origin, get_args, Any, Optional, get_type_hints
)

__all__ = [
    'AbstractParser',
    'new_parser',
    'new_command_parser',
    'parse_args',
    'parse_command_args',
    'set_options',
    'argument', 'as_argument',
    'with_defaults',
    'print_help',
    'as_dict',
    'copy_argument',
    #
    'Argument'
]

T = TypeVar('T')
Nargs = Literal[
    '*', '+', '?', '...'
]
Actions = Literal[
    'store',
    'store_const',
    'store_true',
    'store_false',
    'append',
    'append_const',
    'extend',
    'count',
    'help',
    'version',
    #
    'boolean'
]


class AbstractParser:
    USAGE: str = None
    """parser usage."""

    DESCRIPTION: str = None
    """parser description."""

    EPILOG: str = None
    """parser epilog. Could be override as a method if its content is dynamic-generated."""

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        with_defaults(obj)
        return obj

    @classmethod
    def new_parser(cls, **kwargs) -> argparse.ArgumentParser:
        """create an ``argparse.ArgumentParser``.

        class variable: ``USAGE``, ``DESCRIPTION`` and ``EPILOG`` are used when creation.

        >>> class A(AbstractParser):
        ...     @classmethod
        ...     def new_parser(cls, **kwargs) -> argparse.ArgumentParser:
        ...         return super().new_parser(**kwargs)

        :param kwargs: keyword parameters to ArgumentParser
        :return: an ArgumentParser.
        """
        return new_parser(cls, **kwargs)

    def main(self, args: Union[list[str], tuple[list[str], list[str]]] = None, *,
             exit_on_error=True):
        """parsing the commandline input *args* and set the argument attributes,
        then call :meth:`.run()`.

        **Example**

        if overwrite with the argument default, use *args*

        >>> AbstractParser().main((['--source=allen_mouse_25um', '--region=VISal,VISam,...'], []))

        :param args: commandline arguments, or a tuple of (prepend, append) arguments
        :param exit_on_error: exit when commandline parsed fail. Otherwise, raise a ``RuntimeError``.
        """
        if args is not None:
            if isinstance(args, list):
                pass
            elif isinstance(args, tuple):
                prepend, append = args[0], args[1]
                args = [*prepend, *sys.argv[1:], *append]
            else:
                raise TypeError('')

        parser = self.new_parser(reset=True)

        try:
            result = parser.parse_args(args)
        except SystemExit as e:
            if exit_on_error:
                raise
            else:
                raise RuntimeError() from e
        else:
            set_options(self, result)
            self.post_parsing()
            self.run()

    def run(self):
        """called when all argument attributes are set"""
        pass

    def post_parsing(self):
        """called when all argument attributes are set but before :meth:`.run()`.

        It is used for a common operation for a common option class,
        for example, checking arguments before doing things.
        """
        pass


class Argument(object):
    """Descriptor (https://docs.python.org/3/glossary.html#term-descriptor).
    Carried the arguments pass to ``argparse.ArgumentParser.add_argument``.

    **Creation**

    Use :func:`~neuralib.argp.core.argument()`.

    >>> class Example:
    ...     a: str = argument('-a')

    """

    def __init__(self, *options, group: str = None, **kwargs):
        """

        :param options: options
        :param group: argument group.
        :param kwargs:
        """
        self.attr = None
        self.attr_type = Any
        self.group = group
        self.options = options
        self.kwargs = kwargs

    @property
    def default(self):
        try:
            return self.kwargs['default']
        except KeyError:
            pass

        raise ValueError

    @property
    def const(self):
        try:
            return self.kwargs['const']
        except KeyError:
            pass

        raise ValueError

    @property
    def metavar(self) -> Optional[str]:
        return self.kwargs.get('metavar', None)

    @property
    def choices(self) -> Optional[tuple[str, ...]]:
        return self.kwargs.get('choices', None)

    @property
    def required(self) -> bool:
        return self.kwargs.get('required', False)

    @property
    def help(self) -> Optional[str]:
        return self.kwargs.get('help', None)

    def __set_name__(self, owner: Type, name: str):
        self.attr = name
        self.attr_type = get_type_hints(owner).get(name, Any)

    def __get__(self, instance, owner=None):
        if instance is None:
            if owner is not None:  # ad-hoc for the document building
                self.__doc__ = self.help
            return self
        try:
            return instance.__dict__[f'__{self.attr}']
        except KeyError:
            pass

        raise AttributeError(self.attr)

    def __set__(self, instance, value):
        instance.__dict__[f'__{self.attr}'] = value

    def __delete__(self, instance):
        try:
            del instance.__dict__[f'__{self.attr}']
        except KeyError:
            pass

    def add_argument(self, ap: argparse.ArgumentParser, instance):
        """Add this into `argparse.ArgumentParser`.

        :param ap:
        :param instance:
        :return:
        """
        kwargs = self.complete_kwargs()

        try:
            return ap.add_argument(*self.options, **kwargs, dest=self.attr)
        except TypeError as e:
            if isinstance(instance, type):
                name = instance.__name__
            else:
                name = type(instance).__name__
            raise RuntimeError(f'{name}.{self.attr} : ' + repr(e)) from e

    def complete_kwargs(self) -> dict[str, Any]:
        """infer missing keywords.

        :return: kwargs
        """
        attr_type = self.attr_type
        kwargs = dict(self.kwargs)

        if 'type' not in kwargs:
            if attr_type == bool:
                if 'action' not in kwargs:
                    if kwargs.get('default', False) is False:
                        kwargs['action'] = 'store_true'
                        kwargs.setdefault('default', False)
                    else:
                        kwargs['action'] = 'store_false'
                        kwargs.setdefault('default', True)

            elif get_origin(attr_type) == Literal:
                kwargs.setdefault('choices', get_args(attr_type))

            elif get_origin(attr_type) is Union:
                type_args = get_args(attr_type)
                if len(type_args) == 2 and type_args[1] is type(None):
                    if get_origin(type_args[0]) == Literal:
                        kwargs.setdefault('choices', get_args(type_args[0]))
                    elif callable(type_args[0]):
                        kwargs['type'] = type_args[0]

            elif kwargs.get('action', None) in ['append', 'extend']:
                coll_attr_type = get_origin(attr_type)
                if coll_attr_type == list:
                    kwargs['type'] = get_args(attr_type)[0]
                else:
                    raise RuntimeError(f"cannot infer type. {self.attr} missing keyword type.")
            elif callable(attr_type):
                kwargs['type'] = attr_type

        return kwargs

    @overload
    def with_options(self,
                     option: Union[str, dict[str, str]] = None,
                     *options: str,
                     action: Actions = None,
                     nargs: Union[int, Nargs] = None,
                     const: T = None,
                     default: T = None,
                     type: Union[Type, Callable[[str], T]] = None,
                     choices: Sequence[str] = None,
                     required: bool = None,
                     help: str = None,
                     group: str = None,
                     metavar: str = None) -> 'Argument':
        pass

    def with_options(self, *options, **kwargs) -> 'Argument':
        """Modify or update keyword parameter and return a new argument.

        option flags update rule:

        1. ``()`` : do not update options
        2. ``('-a', '-b')`` : replace options
        3. ``(..., '-c')`` : append options
        4. ``({'-a': '-A'})`` : rename options
        4. ``({'-a': '-A'}, ...)`` : rename options, keep options if not in the dict.

        general form:

        ``() | (dict?, ...?, *str)``

        :param options: change option flags
        :param kwargs: change keyword parameters, use `...` to unset parameter
        :return:
        """
        kw = dict(self.kwargs)
        kw['group'] = self.group
        kw.update(kwargs)

        for k in list(kw.keys()):
            if kw[k] is ...:
                del kw[k]

        if len(self.options) > 0:
            if len(options) == 0:
                return Argument(*self.options, **kw)
            elif options[0] is ...:
                return Argument(*self.options, *options[1:], **kw)
            elif isinstance(options[0], dict):
                if len(options) == 1:
                    return Argument(*self._map_options(options[0], False), **kw)
                if len(options) == 2 and options[1] is ...:
                    return Argument(*self._map_options(options[0], True), **kw)
                if options[1] is ...:
                    return Argument(*self._map_options(options[0], True), *options[2:], **kw)
                else:
                    return Argument(*self._map_options(options[0], False), *options[1:], **kw)
            else:
                return Argument(*options, **kw)

        else:
            if len(options) > 0:
                raise RuntimeError('cannot change positional argument to optional')

            return Argument(**kw)

    def _map_options(self, mapping: dict[str, str], keep: bool) -> list[str]:
        new_opt = []
        for old_opt in self.options:
            try:
                new_opt.append(mapping[old_opt])
            except KeyError:
                if keep:
                    new_opt.append(old_opt)
        return new_opt


@overload
def argument(*options: str,
             action: Actions = None,
             nargs: Union[int, Nargs] = None,
             const: T = None,
             default: T = None,
             type: Union[Type, Callable[[str], T]] = None,
             choices: Sequence[str] = None,
             required: bool = None,
             help: str = None,
             group: str = None,
             metavar: str = None) -> T:
    pass


def argument(*options: str, **kwargs):
    """create an argument attribute.

    Example:

    >>> class Example:
    ...     # create a bool flag
    ...     bool_flag: bool = argument('-f')
    ...     # create a single value option
    ...     str_value: str = argument('-a', metavar='VALUE')
    ...     # create a single value option with type auto-casting
    ...     int_value: int = argument('-i', metavar='VALUE')
    ...     # create a position argument
    ...     pos_value: str = argument(metavar='VALUE')
    ...     # create a multiple value option
    ...     list_value: list[str] = argument('-l', metavar='VALUE', nargs=2, action='append')


    :param kwargs: Please see ``argparse.ArgumentParser.add_argument`` for detailed.
    """
    if not all([it.startswith('-') for it in options]):
        raise RuntimeError(f'options should startswith "-". {options}')
    return Argument(*options, **kwargs)


def as_argument(a) -> Argument:
    """cast argument attribute as an :class:`~neuralib.argp.core.Argument` for type checking framework/IDE."""
    if isinstance(a, Argument):
        return a
    raise TypeError


def foreach_arguments(instance: Union[T, type[T]]) -> Iterable[Argument]:
    """iterating all argument attributes in instance.

    This method will initialize Argument.

    :param instance:
    :return:
    """
    if isinstance(instance, type):
        clazz = instance
    else:
        clazz = type(instance)

    arg_set = set()
    for clz in reversed(clazz.mro()):
        if (ann := getattr(clz, '__annotations__', None)) is not None:
            for attr in ann:
                if isinstance((arg := getattr(clazz, attr, None)), Argument) and attr not in arg_set:
                    arg_set.add(attr)
                    yield arg


def new_parser(instance: Union[T, type[T]], reset=False, **kwargs) -> argparse.ArgumentParser:
    """Create ``ArgumentParser`` for instance.

    :param instance:
    :param reset: reset argument attributes. do nothing if *instance* isn't an instance.
    :param kwargs: keywords for creating :class:`argparse.ArgumentParser`.
    :return:
    """
    if isinstance(instance, AbstractParser) or (isinstance(instance, type) and issubclass(instance, AbstractParser)):
        kwargs.setdefault('usage', instance.USAGE)
        kwargs.setdefault('description', instance.DESCRIPTION)
        kwargs.setdefault('formatter_class', argparse.RawTextHelpFormatter)
        epilog = instance.EPILOG
        if callable(epilog):
            epilog = epilog()
        kwargs.setdefault('epilog', epilog)

    ap = argparse.ArgumentParser(**kwargs)

    gp = {}
    for arg in foreach_arguments(instance):
        if instance is not None and not isinstance(instance, type) and reset:
            arg.__delete__(instance)

        if arg.group is None:
            tp = ap
        elif arg.group in gp:
            tp = gp[arg.group]
        else:
            gp[arg.group] = tp = ap.add_argument_group(arg.group)

        arg.add_argument(tp, instance)

    return ap


def new_command_parser(parsers: dict[str, Union[AbstractParser, type[AbstractParser]]],
                       usage: str = None,
                       description: str = None,
                       reset=False) -> argparse.ArgumentParser:
    """Create ``ArgumentParser`` for :class:`~neuralib.argp.core.AbstractParser` s.

    :param parsers: dict of command to :class:`~neuralib.argp.core.AbstractParser`.
    :param usage: parser usage
    :param description: parser description
    :param reset: reset argument attributes. do nothing if *parsers*'s value isn't an instance.
    :return:
    """
    ap = argparse.ArgumentParser(usage=usage, description=description)
    sp = ap.add_subparsers()

    for cmd, pp in parsers.items():
        ppap = new_parser(pp, reset=reset)
        ppap.set_defaults(main=pp)
        sp.add_parser(cmd, help=pp.DESCRIPTION, parents=[ppap], add_help=False)

    return ap


def set_options(instance: T, result: argparse.Namespace) -> T:
    """set argument attributes from ``argparse.Namespace`` .

    :param instance:
    :param result:
    :return: *instance* itself.
    """
    for arg in foreach_arguments(instance):
        try:
            value = getattr(result, arg.attr)
        except AttributeError:
            pass
        else:
            arg.__set__(instance, value)

    return instance


def parse_args(instance: T, args: list[str] = None) -> T:
    """parsing the commandline input *args* and set the argument attributes.

    :param instance:
    :param args: commandline inputs
    :return:
    """
    return set_options(instance, new_parser(instance, reset=True).parse_args(args))


def parse_command_args(parsers: dict[str, Union[AbstractParser, type[AbstractParser]]],
                       args: list[str] = None,
                       usage: str = None,
                       description: str = None,
                       run_main=True) -> Optional[AbstractParser]:
    """Create ``argparse.ArgumentParser`` for :class:`~neuralib.argp.core.AbstractParser` s.
    Then parsing the commandline input *args* and setting up correspond :class:`~neuralib.argp.core.AbstractParser`.

    :param parsers: dict of command to :class:`~neuralib.argp.core.AbstractParser`.
    :param args: commandline inputs
    :param usage: parser usage
    :param description: parser description.
    :param run_main: run :meth:`~neuralib.argp.core.AbstractParser.run()`
    :return: used :class:`~neuralib.argp.core.AbstractParser`
    """
    ap = new_command_parser(parsers, usage, description, reset=True)
    res = ap.parse_args(args)

    pp: AbstractParser = getattr(res, 'main', None)
    if isinstance(pp, type):
        pp = pp()

    if pp is not None:
        set_options(pp, res)

    if run_main:
        if pp is not None:
            pp.post_parsing()
            pp.run()
        else:
            from rich import print
            print(f'[bold red]should be one of {", ".join(parsers.keys())}')

    return pp


def print_help(instance: T):
    """print help to stdout"""
    new_parser(instance).print_help(sys.stdout)


def with_defaults(instance: T) -> T:
    """Initialize all argument attributes by assign the default value if provided.

    :param instance:
    :return: *instance* itself
    """
    for arg in foreach_arguments(instance):
        ck = arg.complete_kwargs()
        try:
            value = ck['default']
        except KeyError as e:
            if 'action' in ck:
                if ck['action'] == 'store_true':
                    arg.__set__(instance, False)
                else:
                    arg.__set__(instance, True)
            else:
                arg.__set__(instance, None)
        else:
            arg.__set__(instance, value)
    return instance


def as_dict(instance: T) -> dict[str, Any]:
    """collect all argument attributes into a dictionary with attribute name to its value.

    :param instance:
    :return:
    """
    ret = {}
    for arg in foreach_arguments(instance):
        try:
            value = arg.__get__(instance)
        except AttributeError:
            pass
        else:
            ret[arg.attr] = value
    return ret


def copy_argument(opt: T, ref, **kwargs) -> T:
    """copy argument from ref to opt

    :param opt
    :param ref:
    :param kwargs:
    :return:
    """
    shadow = ShadowOption(ref, **kwargs)

    for arg in foreach_arguments(opt):
        try:
            value = getattr(shadow, arg.attr)
        except AttributeError:
            pass
        else:
            # print('set', arg.attr, value)
            arg.__set__(opt, value)
    return opt


class ShadowOption:
    """Shadow options, used to pass wrapped :class:`AbstractOptions`
    """

    def __init__(self, ref, **kwargs):
        self.__ref = ref
        self.__kwargs = kwargs

    def __getattr__(self, attr: str):
        if attr in self.__kwargs:
            return self.__kwargs[attr]

        if attr.startswith('_') and attr[1:] in self.__kwargs:
            return self.__kwargs[attr[1:]]

        if hasattr(self.__ref, attr):
            return getattr(self.__ref, attr)

        raise AttributeError(attr)



"""
Bokeh Base module
=================

:author:
    Ta-Shun Su

[Bokeh](https://docs.bokeh.org/en/latest/) base module for building bokeh application
which is web-based, interactive and python communicate-able.

Simple Example
--------------

>>> from neuralib.bokeh_model import BokehServer, View
>>> class Top(View):
...     def setup(self):
...         from bokeh.models import Div
...         from bokeh.layouts import column
...         return column(Div(text="Hello World!"))
>>> SERVER = BokehServer('Example')
>>> SERVER.start(Top())

General Application Structure
-----------------------------

This structure proposol was following MVC common structure. For this module. **M (Model)** is defined
as a group of functions that handle the data IO and processing. **V (View)** is defined as a group of
classes that hold the bokeh UI components. **C (Controller)** is defined as a group of functions that
bride the Model and View. They are UI event listeners, application controler and other kinds of
function that nothing relate to the data.



>>> # data IO functions
... def list_data() -> list[str]: ...
>>> # Custom ViewComponent
... class DataView(ViewComponent): ...
>>> # Top View
... class Top(View):
...     view_data: DataView
...     def setup(self):
...         # init view_data
...     def update(self):
...         # update view_data
...     # controller functions
...     def on_data_update(self, attr, old, value): ... # property change/update event
...     def on_data_select(self, e): ... # action event
>>> # create cli parser (optional)
>>> # main function (optional)
>>> # create server


"""

import abc
import functools
from typing import Callable, ClassVar, cast, Union, Optional

from bokeh.application import Application
from bokeh.document import Document
from bokeh.models import Model, GlyphRenderer
from bokeh.plotting import figure as _fig
from bokeh.server.server import Server

__all__ = ['Figure', 'View', 'ViewComponent', 'BokehServer']


def Figure(**kwargs):
    return _fig(**kwargs, output_backend='svg')


class View(metaclass=abc.ABCMeta):
    """Top View of bokeh application.

    Example
    -------

    >>> class Top(View):
    ...     # custom viewer, it often follows a Figure
    ...     view_a: ViewA # class ViewA(ViewComponent)
    ...     figure_a: Figure
    ...
    ...     # exported UI components, put here if it needs to be access by controller functions
    ...     select_a: Select
    ...
    ...     def setup(self):
    ...         # initialize custom viewer
    ...         self.figure_a = Figure(...)
    ...         self.view_a = ViewA()
    ...         self.view_a.plot(self.figure_a)
    ...
    ...         # initialize exported UI components
    ...         self.select_a = Select(...)
    ...         self.select_a.on_change('value', on_select_a) # bind 'value' change event listener
    ...
    ...         # local UI components
    ...         # for those simple UIs without complex interaction, such as Button
    ...         button_a = Button('A')
    ...         button_a.on_click(on_btn_a) # bind pressed event listener
    ...
    ...         # layout UI components
    ...         from bokeh.layouts import row, column
    ...         return column(
    ...             self.figure_a,
    ...             row(self.select_a, button_a)
    ...         )

    Implement Note
    --------------

    :class:`View` instance should be singleton, which means there is no second instance existed in the
    same application/python runtime. It is a soft rule (which means you can break this rule without
    and error happen), and this class don't have any mechanism to ensure this rule.

    This class don't force the :meth:`__init__` function signature, which means you can define
    your own `__init__` function with custom parameters, such as arguments from the CLI.
    
    Beside top level updating function :func:`update`, you can define your own updating 
    functions for only smaller group of UI components.

    You can declare all the UI components inside the View class, but you probably will get
    a mess code. Hence, :class:`ViewComponent` is used to group related functions together, included
    some controller functions and data processing functions. By this way, such as variable `view_a`
    in the example, it also provides a namespace that make the code more readable.

    """

    document: Document

    @property
    def title(self) -> str:
        return type(self).__name__

    def get_arg(self, key: str) -> list[str]:
        return list(map(bytes.decode, self.document.session_context.request.arguments[key]))

    @abc.abstractmethod
    def setup(self) -> Model:
        """setup application top view. This function is called by BokehServer only once when
        the server created and opened the web browser.

        In this function, you need to initialize the UI components and return the layout.
        """
        pass

    def update(self):
        """Top level UI components updating function. This function is called by BokehServer
        when :func:`setup` has done.

        In this function, you need to initialize the data and update the UI components' state.
        """
        pass

    def run_later(self, callback: Callable, *args, **kwargs):
        self.document.add_next_tick_callback(functools.partial(callback, *args, **kwargs))

    def run_timeout(self, delay: int, callback: Callable, *args, **kwargs):
        self.document.add_timeout_callback(functools.partial(callback, *args, **kwargs), delay)

    def run_periodic(self, cycle: int, callback: Callable, *args, **kwargs):
        self.document.add_periodic_callback(functools.partial(callback, *args, **kwargs), cycle)

    def on_message(self, message: str, reset: bool):
        """for logging purpose. i.e., TextAreaInput bokeh widget"""
        pass


class ViewComponent(metaclass=abc.ABCMeta):
    """A UI component that provides certain graph on specific data type.

    General Structure
    -----------------

    >>> class MyView(ViewComponent):
    ...     # for ColumnDataSource attribute's name, there is no hard rule
    ...     data_a: ColumnDataSource
    ...     # for GlyphRenderer attribute's name, it should be named with prefix 'render_'
    ...     render_a: GlyphRenderer
    ...     def __init__(self):
    ...         # initialize an empty data
    ...         self.data_a = ColumnDataSource(data=dict(...))
    ...     def plot(self, figure):
    ...         # plotting data
    ...         self.render_a = figure.plot(self.data_a)
    ...     # function to update the render, data
    ...     def update(self, data):
    ...         # update data
    ...         self.data_a.activity = dict(...)


    Implement Note
    --------------

    I suggest make plotting related properties as @property, because they usually need to update/invalid
    other attributes/properties when it is updated. In order to improve the performance, you don't need
    to update the graph/render for each property update in an event loop. Because most render update functions
    have to process data and set value into :class:`ColumnDataSource`, and :class:`ColumnDataSource` need to
    sync the content with the web browser, it will take many times for every property update. You can mention
    (in document) the caller have to call the render update function after several properties update function.

    """

    @abc.abstractmethod
    def plot(self, figure: Figure, **kwargs):
        """plot data in *figure*.

        :param figure: Figure.
        :param kwargs: plotting parameters.
        """
        pass

    @abc.abstractmethod
    def update(self):
        """update the plot"""
        pass

    def set_visible(self, visible: bool, pattern: str = None):
        """Set the visible state of renders for those name contain *pattern*.
        It is a recursive function that it also update the renders inside the
        :class:`ViewComponent` attributes.

        :param visible:
        :param pattern: str in attribute name
        """
        for name in dir(self):
            render = getattr(self, name)
            if name.startswith('render_'):
                if pattern is None or pattern in name:
                    if isinstance(render, list):
                        for _render in render:
                            cast(GlyphRenderer, _render).visible = visible
                    elif isinstance(render, dict):
                        for _render in render.values():
                            cast(GlyphRenderer, _render).visible = visible
                    else:
                        cast(GlyphRenderer, render).visible = visible
            elif isinstance(render, ViewComponent):
                render.set_visible(visible, pattern)

    def list_renders(self, pattern: str = None, recursive: bool = False) -> list[GlyphRenderer]:
        """list all renders for those name contain *pattern*.

        :param pattern: str in attribute name
        :param recursive: recursive find all renders from :class:`ViewComponent` attributes.
        :return:
        """
        ret = []
        for name in dir(self):
            render = getattr(self, name)
            if name.startswith('render_'):
                if pattern is None or pattern in name:
                    if isinstance(render, list):
                        ret.extend(render)
                    elif isinstance(render, dict):
                        ret.extend(render.values())
                    else:
                        ret.append(render)
            elif isinstance(render, ViewComponent) and recursive:
                ret.extend(render.list_renders(pattern, recursive))
        return ret


class BokehServer:
    """Bokeh application server. It is a singleton class, that only one instance can hold the OS port.

    Implement Note
    --------------

    For now, it is a single web-page application, that don't provide any methods to route to other paths.

    """

    INSTANCE: ClassVar['BokehServer'] = None

    server: Server

    def __init__(self, theme: str = 'dark_minimal'):
        self.theme = theme

    def start(self,
              viewer: Union[View, Application, dict[str, Union[View, Application]]],
              open_url: Optional[str] = '/', **kwargs):
        """start serving.

        :param viewer: top view
        :param open_url
        """
        if BokehServer.INSTANCE is None:
            BokehServer.INSTANCE = self
        else:
            raise RuntimeError('server is running')

        if isinstance(viewer, (View, Application)):
            router = {'/': self.page(viewer)}
        else:
            router = {k: self.page(v) for k, v in viewer.items()}

        self.server = Server(router, num_procs=1, **kwargs)
        self.server.start()

        if open_url is not None:
            self.server.io_loop.add_callback(self.server.show, open_url)

        self.server.run_until_shutdown()

    def page(self, viewer: Union[View, Application]):
        """ """
        if isinstance(viewer, Application):
            return viewer

        def _page(doc: Document):
            self._init_doc(doc)
            viewer.document = doc
            doc.add_root(viewer.setup())
            doc.title = viewer.title
            doc.add_next_tick_callback(viewer.update)

        return _page

    def _init_doc(self, doc: Document):
        doc.theme = self.theme



from __future__ import annotations

from pathlib import Path
from typing import ClassVar, TypedDict

import allensdk.core.structure_tree
from allensdk.core.reference_space_cache import ReferenceSpaceCache

from neuralib.util.io import ALLEN_SDK_DIRECTORY
from neuralib.util.util_type import PathLike

__all__ = ['AllenReferenceWrapper']


class StructureTreeDict(TypedDict):
    """allen structure tree dict"""
    acronym: str
    graph_id: int
    graph_order: int
    id: int
    name: str
    structure_id_path: list[int]
    structure_set_ids: list[int]
    rgb_triplet: list[int]


class AllenReferenceWrapper:
    """Class for load the data in allen"""

    REFERENCE_SPACE_KEY: ClassVar[str] = 'allensdk'
    STRUCTURE_GRAPH_ID: ClassVar[int] = 1

    def __init__(self,
                 resolution: int = 10,
                 output: PathLike | None = None):
        """
        :param resolution: resolution in um
        :param output: output directory for downloading cache. By default, save in ``CCF_CACHE_DIRECTORY``
        """

        if output is None:
            output = ALLEN_SDK_DIRECTORY

        if not output.exists():
            output.mkdir(parents=True, exist_ok=True)

        self.source_root = Path(output)
        self.reference = ReferenceSpaceCache(resolution,
                                             self.REFERENCE_SPACE_KEY,
                                             manifest=self.source_root / 'manifest.json')

    def structure_tree(self) -> allensdk.core.structure_tree.StructureTree:
        return self.reference.get_structure_tree(structure_graph_id=self.STRUCTURE_GRAPH_ID)

    def get_structures_by_name(self, name: list[str]) -> list[StructureTreeDict]:
        return self.structure_tree().get_structures_by_name(name)



"""
SQL statement builder
=====================

:author:
    Ta-Shun Su

This module provide a Python functions to build a SQL (sqlite3) statement.
"""
from . import util
from .annotation import *
from .cli import *
from .connection import *
from .func import *
from .func_date import *
from .func_stat import *
from .func_win import *
from .literal import *
from .stat_start import *
from .table_nt import *



from __future__ import annotations

import colorsys
from typing import TypedDict

import numpy as np

from neuralib.plot import plot_figure
from neuralib.util.util_type import PathLike

__all__ = [
    'plot_task_gantt',
    'draw_circular_annotation',
    'generate_dots',
    'draw_random_dots'
]


class GanttDict(TypedDict, total=False):
    task: str
    start: str
    finish: str
    resource: str


def plot_task_gantt(jobs: list[GanttDict]):
    """
    plot gantt
    * Example

    >>> x = [GanttDict(task='task1', start='2023-09-01', finish='2023-11-30'),
    ...      GanttDict(task='task2', start='2024-01-01', finish='2024-09-30')]
    >>> plot_task_gantt(x)

    :param jobs:
    :return:
    """
    import pandas as pd
    import plotly.express as px

    df = pd.DataFrame(jobs)
    if 'resource' not in df.columns:
        color = 'task'
    else:
        color = 'resource'
    fig = px.timeline(df, x_start='start', x_end='finish', y='task', color=color)
    fig.update_layout(
        width=1200,
        height=400
    )
    fig.show()


# ================= #

def draw_circular_annotation(major_radius: float = 3,
                             minor_radius: float = 2,
                             width: float = 10,
                             output: PathLike | None = None):
    """plot circular color annotation. i.e., used in retinotopic mapping illustration"""
    center_x, center_y = 0, 0
    n_points = 1000
    theta = np.linspace(0, 2 * np.pi, n_points)

    x = center_x + major_radius * np.cos(theta)
    y = center_y + minor_radius * np.sin(theta)

    hues = np.linspace(0, 1, n_points)
    colors = [colorsys.hsv_to_rgb(hue, 1, 1) for hue in hues]

    with plot_figure(output) as ax:
        for p in range(n_points - 1):
            ax.plot(x[p:p + 2], y[p:p + 2], color=colors[p], linewidth=width)

        ax.set_aspect('equal', adjustable='datalim')

        ax.set_xlim(-major_radius - 1, major_radius + 1)
        ax.set_ylim(-minor_radius - 1, minor_radius + 1)
        ax.axis('off')


# ================= #

def generate_dots(n_dots: int = 20,
                  min_distance: float = 0.1,
                  border_distance: float = 0.1):
    """

    :param n_dots:
    :param min_distance: minimum distance to avoid overlap
    :param border_distance:  minimum distance from the border
    :return:
    """
    ret = []
    existing_dots = []
    while len(ret) < n_dots:
        x = np.random.uniform(border_distance, 1 - border_distance)
        y = np.random.uniform(border_distance, 1 - border_distance)

        # Check for overlap with existing coordinates
        overlap = False
        for coord in existing_dots:
            if np.sqrt((x - coord[0]) ** 2 + (y - coord[1]) ** 2) < min_distance:
                overlap = True
                break

        if not overlap:
            ret.append((x, y))
            existing_dots.append((x, y))

    return ret


def draw_random_dots(n_dots: int = 20,
                     min_distance: float = 0.1,
                     border_distance: float = 0.01,
                     output: PathLike | None = None,
                     pixels: tuple[int, int] = (1000, 1000)):
    dots = generate_dots(n_dots, min_distance, border_distance)

    dpi = 500
    with plot_figure(output,
                     figsize=(pixels[0] / dpi, pixels[1] / dpi), dpi=dpi) as ax:
        for d in dots:
            ax.scatter(d[0], d[1], color='black', s=20)

        ax.set(xlim=(0, 1), ylim=(0, 1))
        ax.set_facecolor('white')
        ax.axis('off')
        ax.set_aspect('equal', adjustable='box')



from __future__ import annotations

import dataclasses
import json
from dataclasses import dataclass, field
from pathlib import Path
from pprint import pprint
from typing import TypeVar, Generic, TypedDict

import numpy as np
from scipy.io import loadmat

from neuralib.util.json import JsonEncodeHandler
from neuralib.util.util_type import PathLike

__all__ = [
    'SBXInfo',
    'sbx_to_json',
    #
    'screenshot_to_tiff'
]

T = TypeVar('T')  # matstruct
A = TypeVar('A')  # sbx attr


@dataclass(frozen=True)
class SBXInfo(Generic[T, A]):
    """for each recording session, `exp.mat` from scanbox"""
    scanbox_version: str
    objective: str
    abort_bit: int
    area_line: int
    ballmotion: np.ndarray
    bytesPerBuffer: int
    scanmode: int
    config: ConfigInfo
    calibration: list[CalibrationInfo]
    channels: int
    messages: np.ndarray
    opto2pow: np.ndarray
    otparam: np.ndarray  # ETL setting?
    otwave: np.ndarray
    otwave_um: np.ndarray
    otwavestyle: int
    postTriggerSamples: int
    power_depth_link: int
    recordsPerBuffer: int
    resfreq: int
    sz: np.ndarray  # line/frames?
    usernotes: np.ndarray
    volscan: int  # volumetric img? bool?

    # opt
    nchan: int | None

    @property
    def magidx(self) -> int:
        """info.config.magnification, used for `CalibrationInfo` idx"""
        return self.config.magnification - 1

    def print_asdict(self) -> None:
        from dataclasses import asdict
        pprint(asdict(self))

    @staticmethod
    def _try_find_attr(obj: T, attr: str) -> A | list:
        try:
            ret: A = getattr(obj, attr)
        except AttributeError:
            return []
        else:
            return ret

    # =============================== #
    # attributes from SBX .mat output #
    # =============================== #

    @classmethod
    def load(cls, root: PathLike) -> SBXInfo:
        info = loadmat(root, squeeze_me=True, struct_as_record=False)['info']

        try:
            nchan = info.chan.nchan  # version >= 3
        except AttributeError:
            nchan = None

        return SBXInfo(
            scanbox_version=str(info.scanbox_version),
            objective=info.objective,
            abort_bit=info.abort_bit,
            area_line=info.area_line,
            ballmotion=info.ballmotion,
            bytesPerBuffer=info.bytesPerBuffer,
            scanmode=info.scanmode,
            config=cls.get_config_info(info.config),
            calibration=cls.get_calibration_info_list(info.calibration),
            channels=info.channels,
            messages=info.messages,
            opto2pow=info.opto2pow,
            otparam=info.otparam,
            otwave=info.otwave,
            otwave_um=info.otwave_um,
            otwavestyle=info.otwavestyle,
            postTriggerSamples=info.postTriggerSamples,
            power_depth_link=info.power_depth_link,
            recordsPerBuffer=info.recordsPerBuffer,
            resfreq=info.resfreq,
            sz=info.sz,
            usernotes=info.usernotes,
            volscan=info.volscan,
            nchan=nchan
        )

    @dataclass(frozen=True)
    class CalibrationInfo:
        """attr from calibration"""
        x: float
        y: float
        gain_resonant_mult: int
        uv: list = field(default_factory=list)
        delta: list = field(default_factory=list)

    @classmethod
    def get_calibration_info_list(cls, calibration_list) -> list[CalibrationInfo]:
        return [cls.CalibrationInfo(
            x=calinfo.x,
            y=calinfo.y,
            gain_resonant_mult=calinfo.gain_resonant_mult,
            uv=calinfo.uv,
            delta=calinfo.delta
        ) for calinfo in calibration_list]

    # noinspection PyUnresolvedReferences
    @dataclass(frozen=True)
    class ConfigInfo:
        """attr from config"""
        agc: 'AgcInfo'
        # calibration: np.ndarray
        coord_abs: np.ndarray
        coord_rel: np.ndarray
        frame_times: np.ndarray
        frames: int  # total frames
        host_name: str  # BSTATION6
        knobby: 'KnobbyInfo'
        laser_power: float
        laser_power_perc: str  # 75%
        lines: int  # 528
        magnification: int  # idx from 1 in magnification_list
        magnification_list: np.ndarray
        objective: 'ObjectiveInfo'  # directly get useful attr `name`
        objective_type: int
        pmt0_gain: float  # green channel
        pmt1_gain: float  # red channel
        wavelength: int  # laser wavelength. i.e., 920 nm

    @classmethod
    def get_config_info(cls, config: T) -> ConfigInfo:
        return cls.ConfigInfo(
            agc=cls.get_agc_info(config.agc),
            # calibration=config.calibration,
            coord_abs=config.coord_abs,
            coord_rel=config.coord_rel,
            frame_times=config.frame_times,
            frames=config.frames,
            host_name=cls._try_find_attr(config, 'host_name'),
            knobby=cls.get_knobby_info(config.knobby),
            laser_power=config.laser_power,
            laser_power_perc=config.laser_power_perc,
            lines=config.lines,
            magnification=config.magnification,
            magnification_list=config.magnification_list,
            objective=cls.get_objective_info(config.objective),
            objective_type=config.objective_type,
            pmt0_gain=config.pmt0_gain,
            pmt1_gain=config.pmt1_gain,
            wavelength=config.wavelength,
        )

    @dataclass(frozen=True)
    class AgcInfo:
        """attr from info.config.agc"""
        agc_prctile: np.ndarray
        enable: int
        threshold: int

    @classmethod
    def get_agc_info(cls, agc: T) -> AgcInfo:
        return cls.AgcInfo(
            agc_prctile=agc.agc_prctile,
            enable=agc.enable,
            threshold=agc.threshold
        )

    # noinspection PyUnresolvedReferences
    @dataclass(frozen=True)
    class KnobbyInfo:
        """attr from info.config.knobby"""
        pos: 'KnobbyPosInfo'
        schedule: np.ndarray

    @classmethod
    def get_knobby_info(cls, knobby: T) -> KnobbyInfo:
        return cls.KnobbyInfo(
            pos=cls.get_knobby_pos_info(knobby.pos),
            schedule=knobby.schedule
        )

    @dataclass(frozen=True)
    class KnobbyPosInfo:
        """attr from info.config.knobby.pos
        manipulator coordinates
        """
        a: float
        x: float
        y: float
        z: float

    @classmethod
    def get_knobby_pos_info(cls, pos: T) -> KnobbyPosInfo:
        return cls.KnobbyPosInfo(
            a=pos.x,
            x=pos.x,
            y=pos.y,
            z=pos.z
        )

    @dataclass(frozen=True)
    class ObjectiveInfo:
        name: str

    @classmethod
    def get_objective_info(cls, objective: T) -> ObjectiveInfo:
        return cls.ObjectiveInfo(
            name=objective.name
        )

    # =============================== #
    # attributes from SBX .mat output #
    # =============================== #

    @property
    def fov_distance(self) -> tuple[float, float]:
        """(X, Y) in um.

        Note this value might be hardware dependent. value return is internal usage for the lab
        """
        lines = self.get_config_info(self.config).lines

        if self.objective == 'Nikon 16x/0.8w/WD3.0':
            obj = 16
        else:
            raise NotImplementedError('')

        zoom = float(self.get_config_info(self.config).magnification_list[self.magidx])

        return _get_default_scanbox_fov_dimension(lines, obj, zoom)

    def _validate_fov_distance(self) -> bool:
        """due to this is the info only seen in GUI"""
        obj_type = self.objective == 'Nikon 16x/0.8w/WD3.0'
        n_line = self.get_config_info(self.config).lines == 528
        mag = self.get_config_info(self.config).magnification_list[self.magidx] == '1.7'
        return all([obj_type, n_line, mag])


def _get_default_scanbox_fov_dimension(lines: int,
                                       obj_type: int,
                                       zoom: float) -> tuple[float, float]:
    """
    Hardware/settings dependent fov size according to recording configuration

    :param lines: number of lines for the scanning fov
    :param obj_type: objective magnification. i.e., 16X
    :param zoom: zoom setting during acquisition
    :return: (X, Y) in um
    """
    # ~ 30hz
    if lines == 528 and obj_type == 16:
        if zoom == 1:
            return 1396, 1056
        elif zoom == 1.2:
            return 1284, 978
        elif zoom == 1.4:
            return 1023, 765
        elif zoom == 1.7:
            return 892, 667
        elif zoom == 2.0:
            return 716, 531
        elif zoom == 2.4:
            return 632, 463
        else:
            raise NotImplementedError('check scanbox GUI directly')

    else:
        raise NotImplementedError('check scanbox GUI directly')


def sbx_to_json(matfile: Path | str,
                output: Path | None = None,
                verbose: bool = True) -> None:
    """
    save .mat scanbox output file as json file

    :param matfile: .mat filepath
    :param output: output filepath, if None, create a json file in the same directory
    :param verbose: pprint as dict
    :return:
    """
    if isinstance(matfile, str):
        matfile = Path(matfile)

    mat = SBXInfo.load(matfile)
    if output is None:
        output = matfile.with_name('sbx').with_suffix('.json')

    dy = dataclasses.asdict(mat)

    if verbose:
        pprint(dy)

    with open(output, "w") as outfile:
        json.dump(dy, outfile, sort_keys=True, indent=4, cls=JsonEncodeHandler)


# ========== #
# ScreenShot #
# ========== #

class SBXScreenShot(TypedDict):
    __header__: str
    __version__: str
    __globals__: str
    img: np.ndarray


def screenshot_to_tiff(mat_file: PathLike,
                       output: PathLike | None = None) -> None:
    """
    Scanbox screenshot result (.mat) to tif file
    :param mat_file:
    :param output: save output path, otherwise show
    :return:
    """
    dat: SBXScreenShot = loadmat(mat_file)
    img = dat['img']

    if output is None:
        import matplotlib.pyplot as plt
        plt.imshow(img)
        plt.show()
    else:
        import tifffile
        tifffile.imwrite(output, img)



from __future__ import annotations

from pathlib import Path
from typing import TypedDict

import attrs
import numpy as np
from typing_extensions import Self

from neuralib.util.util_verbose import fprint

__all__ = ['UserCluster',
           'RasterOptions',
           'RasterMapResult']


class UserCluster(TypedDict, total=False):
    """GUI selected clusters"""

    ids: np.ndarray
    """neuronal ids"""
    slice: slice
    """binned neurons range"""
    binsize: int
    """neuron bins"""
    color: np.ndarray


class RasterOptions(TypedDict, total=False):
    """Run Rastermap model options"""

    n_clusters: int
    """number of clusters to compute. TODO check how affect the result"""
    n_PCs: int
    """number of PCs to use"""
    time_lag_window: float
    """use future timepoints to compute correlation"""
    locality: float
    """locality in sorting to find sequences"""
    n_splits: int
    """TODO"""
    time_bin: int
    """TODO"""
    grid_upsample: int
    """default value, 10 is good for large recordings"""

    mean_time: bool
    verbose: bool
    verbose_sorting: bool
    end_time: int
    start_time: int


@attrs.define
class RasterMapResult:
    """Container for storing the rastermap result,
    For both GUI load and customized plotting purpose

    `Dimension parameters`:

        N = number of neurons

        T = number of image pulse

        C = number of clusters = N / binsize

    """

    filename: str
    """neural activity filename (N, T)"""
    save_path: str
    """filename for the rastermap result save"""
    isort: np.ndarray
    """(N,)"""
    embedding: np.ndarray
    """(N,1)"""
    ops: RasterOptions
    """`RasterOptions`"""
    user_clusters: list[UserCluster] = attrs.field(default=attrs.Factory(list))
    """list of clusters `UserCluster`"""

    super_neurons: np.ndarray | None = attrs.field(default=None)
    """super neuron activity (C, T)"""

    @classmethod
    def load(cls, path: Path) -> Self:
        """
        Load the results from rastermap output

        :param path: file path of the rastermap output
        :return: :class:`RasterMapResult`
        """
        dat = np.load(path, allow_pickle=True).item()
        fprint(f'LOAD ->{path}', vtype='io')

        return RasterMapResult(
            filename=dat['filename'],
            save_path=dat['save_path'],
            isort=dat['isort'],
            embedding=dat['embedding'],
            ops=dat['ops'],
            user_clusters=dat['user_clusters'],
            super_neurons=dat['super_neurons']
        )

    def save(self, path: Path) -> None:
        """For GUI loading & cache computing for plotting in different time domains"""
        proc = {
            'filename': self.filename,
            'save_path': self.save_path,
            'isort': self.isort,
            'embedding': self.embedding,
            'ops': self.ops,
            'user_clusters': self.user_clusters,
            'super_neurons': self.super_neurons,
        }

        np.save(path, proc, allow_pickle=True)
        fprint(f'SAVE ->{path}', vtype='io')

    @property
    def n_super(self) -> int:
        """n_clusters"""
        return len(self.super_neurons)



"""
Module provide basic function for sending Slack messages with Python
=====================================================================

.. seealso:: `<https://www.datacamp.com/tutorial/how-to-send-slack-messages-with-python>`_

1. Configure a Slack App
-------------------------

    1.1. Create an app from https://api.slack.com/app

    1.2. Find the "OAuth & Permissions" tab to open access (i.e., `chat:write`, `chat:write.customize`, `files:read`, `files:write`)

    1.3. Click "Install to Workspace"

    1.4. See token, which starts with ``xoxb-``

    1.5. In Slack, find the ``Apps`` tab on the left side to add the created app

2. Configure an env file
-------------------------

    2.1. Put it in the repo as a (*.env file), **DO NOT** control it with GIT (add to .gitignore)

    2.2. The env file should contain two keys: ``SLACK_TOKEN`` (xoxb-*) and ``CHANNEL_ID`` to send the message

3. Call the function
---------------------

.. code-block:: python

    from neuralib.tools import send_slack_message
    env_file = ...  # env file path
    send_slack_message(env_file, 'Hello, slack!')



"""
import os

from slack import WebClient
from slack.errors import SlackApiError

from neuralib.util.util_type import PathLike
from neuralib.util.util_verbose import fprint

__all__ = ['send_slack_message']


def send_slack_message(env_file: PathLike,
                       message: str) -> None:
    """
    Send message to slack channel

    :param env_file: env file with fields ``SLACK_TOKEN`` (i.e., SLACK_TOKEN=xoxb-<USER_SLACK_TOKEN>),
        and ``CHANNEL_ID`` (i.e., ``#general`` or ``ID``)
    :param message: message to send
    :return:
    """

    try:
        import slack
        from dotenv import load_dotenv
        load_dotenv(dotenv_path=env_file)

        try:
            client = WebClient(token=os.environ['SLACK_TOKEN'])
        except KeyError:
            fprint('To get the slack notification, generate token in local machine first', vtype='warning')
        else:
            try:
                client.chat_postMessage(
                    channel=os.environ['CHANNEL_ID'],
                    text=message,
                )
            except SlackApiError as e:
                print(f"Error sending message: {e.response['error']}")

    except BaseException as e:  # not care, internet disconnected
        fprint(e, vtype='error')



from __future__ import annotations

import abc
import contextlib
import datetime
from collections.abc import Sequence, Iterable
from typing import overload, Any, TYPE_CHECKING, Literal, ContextManager, TypeVar, Generic, Union, Optional

from typing_extensions import Self

from .table import Field

if TYPE_CHECKING:
    from .stat import SqlStat, SqlSelectStat

__all__ = [
    'SqlExpr',
    'wrap', 'wrap_seq',
    'SqlLiteral', 'SqlPlaceHolder', 'SqlField', 'SqlAlias', 'SqlAliasField', 'SqlSubQuery',
    'SqlOper', 'SqlOrderOper', 'SqlCastOper', 'SqlCompareOper', 'SqlUnaryOper', 'SqlBinaryOper', 'SqlFuncOper',
    'SqlVarArgOper', 'SqlConcatOper', 'SqlExistsOper', 'SqlAggregateFunc', 'SqlWindowDef', 'SqlWindowFunc',
    'SqlCaseExpr', 'SqlCteExpr'
]


class SqlExpr(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def __sql_stat__(self, stat: SqlStat):
        pass

    def __matmul__(self, other: str) -> SqlAlias[Self]:
        return SqlAlias(self, other)

    # ============== #
    # equal operator #
    # ============== #

    def __eq__(self, other) -> SqlCompareOper:
        return SqlCompareOper('=', self, wrap(other))

    def __ne__(self, other) -> SqlCompareOper:
        return SqlCompareOper('!=', self, wrap(other))

    def is_null(self) -> SqlCompareOper:
        return SqlCompareOper('IS', self, wrap(None))

    def is_not_null(self) -> SqlCompareOper:
        return SqlCompareOper('IS NOT', self, wrap(None))

    def like(self, value: str) -> SqlCompareOper:
        if value is not None and not isinstance(value, str):
            raise TypeError()

        oper = 'LIKE'
        if value is not None:
            if '%' not in value:
                value += '%'
        return SqlCompareOper(oper, self, SqlLiteral(repr(value)))

    def not_like(self, value: str) -> SqlCompareOper:
        if value is not None and '%' not in value:
            value += '%'
        return SqlCompareOper('NOT LIKE', self, SqlLiteral(repr(value)))

    def glob(self, value: str) -> SqlCompareOper:
        if value is not None and not isinstance(value, str):
            raise TypeError()

        return SqlCompareOper('GLOB', self, SqlLiteral(repr(value)))

    def __contains__(self, item) -> SqlCompareOper:
        from .stat import SqlStat

        if isinstance(item, str):
            return self.like('%' + item + '%')
        elif isinstance(item, (range, slice)):
            return self.between(item)
        elif isinstance(item, (list, tuple, SqlStat)):
            return self.contains(item)
        else:
            raise TypeError()

    @overload
    def between(self, value: Any, value2: Any) -> SqlCompareOper:
        pass

    @overload
    def between(self, value: Union[tuple[Any, Any], range, slice]) -> SqlCompareOper:
        pass

    def between(self, value, value2=None) -> SqlCompareOper:
        if value2 is None:
            return SqlCompareOper('BETWEEN', self, value)
        else:
            return SqlCompareOper('BETWEEN', self, (wrap(value), wrap(value2)))

    @overload
    def not_between(self, value: Any, value2: Any) -> SqlCompareOper:
        pass

    @overload
    def not_between(self, value: Union[tuple[Any, Any], range, slice]) -> SqlCompareOper:
        pass

    def not_between(self, value, value2=None) -> SqlCompareOper:
        if value2 is None:
            return SqlCompareOper('NOT BETWEEN', self, value)
        else:
            return SqlCompareOper('NOT BETWEEN', self, (wrap(value), wrap(value2)))

    def contains(self, value: Union[Sequence, SqlStat]) -> SqlCompareOper:
        from .stat import SqlStat

        if isinstance(value, (tuple, list)):
            return SqlCompareOper('IN', self, tuple(value))
        elif isinstance(value, SqlStat):
            return SqlCompareOper('IN', self, wrap(value))
        else:
            raise TypeError()

    def not_contains(self, value: Union[Sequence, SqlStat]) -> SqlCompareOper:
        from .stat import SqlStat
        if isinstance(value, (tuple, list)):
            return SqlCompareOper('NOT IN', self, tuple(value))
        elif isinstance(value, SqlStat):
            return SqlCompareOper('NOT IN', self, value)
        else:
            raise TypeError()

    # ================ #
    # compare operator #
    # ================ #

    def __lt__(self, other) -> SqlCompareOper:
        return SqlCompareOper('<', self, wrap(other))

    def __le__(self, other) -> SqlCompareOper:
        return SqlCompareOper('<=', self, wrap(other))

    def __gt__(self, other) -> SqlCompareOper:
        return SqlCompareOper('>', self, wrap(other))

    def __ge__(self, other) -> SqlCompareOper:
        return SqlCompareOper('>=', self, wrap(other))

    # =============== #
    # and/or operator #
    # =============== #

    def __and__(self, other: Union[str, SqlExpr]) -> SqlExpr:
        return SqlVarArgOper('AND', [self, wrap(other)])

    def __or__(self, other: Union[str, SqlExpr]) -> SqlExpr:
        return SqlVarArgOper('OR', [self, wrap(other)])

    # =============== #
    # number operator #
    # =============== #

    def __lshift__(self, other: Union[int, SqlExpr]) -> SqlExpr:
        return SqlBinaryOper('<<', self, wrap(other))

    def __rshift__(self, other: Union[int, SqlExpr]) -> SqlExpr:
        return SqlBinaryOper('>>', self, wrap(other))

    def __invert__(self) -> SqlExpr:
        return SqlUnaryOper("~", self)

    def __pos__(self) -> SqlExpr:
        return SqlUnaryOper("+", self)

    def __neg__(self) -> SqlExpr:
        return SqlUnaryOper("-", self)

    def __add__(self, other: Union[int, float, SqlExpr]) -> SqlExpr:
        return SqlBinaryOper("+", self, wrap(other))

    def __radd__(self, other: Union[int, float]) -> SqlExpr:
        return SqlBinaryOper("+", wrap(other), self)

    def __sub__(self, other: Union[int, float, SqlExpr]) -> SqlExpr:
        return SqlBinaryOper("-", self, wrap(other))

    def __rsub__(self, other: Union[int, float]) -> SqlExpr:
        return SqlBinaryOper("-", wrap(other), self)

    def __mul__(self, other: Union[int, float, SqlExpr]) -> SqlExpr:
        return SqlBinaryOper("*", self, wrap(other))

    def __rmul__(self, other: Union[int, float]) -> SqlExpr:
        return SqlBinaryOper("*", wrap(other), self)

    def __truediv__(self, other: Union[int, float, SqlExpr]) -> SqlExpr:
        return SqlBinaryOper("/", self, wrap(other))

    def __rtruediv__(self, other: Union[int, float]) -> SqlExpr:
        return SqlBinaryOper("/", wrap(other), self)

    def __mod__(self, other: Union[int, float, SqlExpr]) -> SqlExpr:
        return SqlBinaryOper('%', self, wrap(other))

    def __rmod__(self, other: Union[int, float]) -> SqlExpr:
        return SqlBinaryOper("%", wrap(other), self)

    # ================= #
    # spectral operator #
    # ================= #

    def cast(self, t: type) -> SqlExpr:
        if t in (bool,):
            return SqlBinaryOper('CAST', self, 'BOOLEAN')
        elif t in (int,):
            return SqlBinaryOper('CAST', self, 'INT')
        elif t in (float,):
            return SqlBinaryOper('CAST', self, 'FLOAT')
        elif t in (str,):
            return SqlBinaryOper('CAST', self, 'TEXT')
        else:
            raise TypeError()


def wrap(other: Any) -> SqlExpr:
    from .stat import SqlStat

    if other is None:
        return SqlLiteral.SQL_NULL
    elif other is True:
        return SqlLiteral.SQL_TRUE
    elif other is False:
        return SqlLiteral.SQL_FALSE
    elif isinstance(other, str) and other.startswith(':'):
        return SqlLiteral(other)
    elif isinstance(other, (int, float, str)):
        return SqlPlaceHolder(other)
    elif isinstance(other, SqlExpr):
        return other
    elif isinstance(other, SqlStat):
        return SqlSubQuery(other)
    elif isinstance(other, datetime.date):
        return SqlLiteral(repr(other.strftime('%Y-%m-%d')))
    elif isinstance(other, datetime.datetime):
        return SqlLiteral(repr(other.strftime('%Y-%m-%d %H:%M:%S')))
    else:
        raise TypeError(repr(other))


def wrap_seq(*other) -> list[SqlExpr]:
    return list(map(wrap, other))


def sql_join(stat: SqlStat, sep: str, exprs: Iterable[SqlExpr]):
    first = True
    for expr in exprs:
        if not first:
            stat.add(sep)
        else:
            first = False

        expr.__sql_stat__(stat)


def sql_join_set(stat: SqlStat, sep: str, exprs: Iterable[Union[bool, SqlCompareOper]]):
    first = True
    for expr in exprs:
        if not first:
            stat.add(sep)
        else:
            first = False

        if isinstance(expr, SqlCompareOper) and expr.oper == '=' and isinstance(expr.left, SqlField):
            field = expr.left.field
            if isinstance(expr.right, SqlPlaceHolder):
                stat.add([field.name, '=', '?'], expr.right.value)
            elif isinstance(expr.right, SqlLiteral):
                stat.add([field.name, '=', '?'], expr.right.value)
            elif isinstance(expr.right, SqlExpr):
                stat.add([field.name, '=', '('])
                expr.right.__sql_stat__(stat)
                stat.add(')')
            else:
                raise TypeError(repr(expr))
        else:
            raise TypeError(repr(expr))


class SqlLiteral(SqlExpr):
    __slots__ = 'value',

    SQL_NULL: 'SqlLiteral'
    SQL_TRUE: 'SqlLiteral'
    SQL_FALSE: 'SqlLiteral'

    def __init__(self, value: str):
        self.value = value

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self.value)

    def __repr__(self):
        return self.value


SqlLiteral.SQL_NULL = SqlLiteral('NULL')
SqlLiteral.SQL_TRUE = SqlLiteral('TRUE')
SqlLiteral.SQL_FALSE = SqlLiteral('FALSE')


class SqlPlaceHolder(SqlExpr):
    __slots__ = 'value',

    def __init__(self, value):
        self.value = value

    def __sql_stat__(self, stat: SqlStat):
        stat.add('?', self.value)

    def __repr__(self):
        return f'?=[{self.value}]'


E = TypeVar('E', bound=SqlExpr)


class SqlAlias(SqlExpr, Generic[E]):
    __slots__ = '_value', '_name'

    def __init__(self, value: Union[E, type[E]], name: str):
        self._value = value
        self._name = name

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self._name)

    def __getattr__(self, item: str) -> SqlAliasField:
        if isinstance(self._value, type):
            from .table import table_field
            table_field(self._value, item)
            return SqlAliasField(self._value, self._name, item)
        elif isinstance(self._value, SqlSubQuery):
            return SqlAliasField(self._value.stat, self._name, item)
        else:
            raise TypeError(f'{self._value}.{item}')


class SqlAliasField(SqlExpr):
    __slots__ = 'table', 'name', 'attr'
    __match_args__ = 'table', 'name', 'attr'

    def __init__(self, table: Union[type, SqlStat, SqlCteExpr], name: str, attr: str):
        self.table = table
        self.name = name
        self.attr = attr

    def __sql_stat__(self, stat: SqlStat):
        stat.add(f'{self.name}.{self.attr}')

    def __str__(self):
        return f'{self.name}.{self.attr}'

    def __repr__(self):
        return f'{self.name}.{self.attr}'


class SqlField(SqlExpr):
    __slots__ = 'field',

    def __init__(self, field: Field):
        self.field = field

    def __sql_stat__(self, stat: SqlStat):
        stat.add(f'{self.table_name}.{self.field.name}')

    @property
    def table(self) -> type:
        return self.field.table

    @property
    def table_name(self) -> str:
        return self.field.table_name

    @property
    def name(self) -> str:
        return self.field.name

    def __repr__(self):
        return f'{self.field.table_name}.{self.field.name}'


class SqlSubQuery(SqlExpr):
    """https://www.sqlite.org/lang_expr.html#subquery_expressions"""

    __slots__ = 'stat',

    def __init__(self, stat: SqlStat):
        self.stat = stat
        stat._connection = None

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self.stat)

    def __repr__(self):
        return '(' + self.stat.build() + ')'


class SqlOper(SqlExpr, metaclass=abc.ABCMeta):
    __slots__ = 'oper',

    def __init__(self, oper: str):
        self.oper = oper


class SqlCompareOper(SqlOper):
    __slots__ = 'oper', 'left', 'right'

    def __init__(self, oper: str, left: SqlExpr, right: Any):
        super().__init__(oper)
        self.left = left
        self.right = right

    def __sql_stat__(self, stat: SqlStat):
        if self.oper in ('IN', 'NOT IN'):
            self.left.__sql_stat__(stat)
            stat.add(self.oper)
            if isinstance(self.right, tuple):
                stat.add(['(', ','.join(map(repr, self.right)), ')'])
            else:
                self.right.__sql_stat__(stat)

        elif self.oper in ('BETWEEN', 'NOT BETWEEN'):
            self.left.__sql_stat__(stat)
            stat.add(self.oper)
            if isinstance(right := self.right, slice):
                a = right.start
                b = right.stop
                if a is None or b is None:
                    raise ValueError()
                stat.add([str(a), 'AND', str(b - 1)])
            elif isinstance(right, range):
                a = right.start
                b = right.stop
                stat.add([str(a), 'AND', str(b - 1)])
            else:
                for i, it in enumerate(right):
                    if i == 1:
                        stat.add('AND')
                    elif i > 1:
                        raise RuntimeError()

                    if isinstance(it, datetime.date):
                        stat.add(it.strftime('%Y-%m-%d'))
                    elif isinstance(it, datetime.datetime):
                        stat.add(it.strftime('%Y-%m-%d %H:%M:%S'))
                    elif isinstance(it, SqlExpr):
                        it.__sql_stat__(stat)
                    else:
                        raise TypeError()
        else:
            self.left.__sql_stat__(stat)
            stat.add(self.oper)
            self.right.__sql_stat__(stat)

    def __repr__(self):
        return '(' + repr(self.left) + self.oper + repr(self.right) + ')'

    def __invert__(self) -> SqlCompareOper:
        return self.not_()

    def not_(self) -> SqlCompareOper:
        if self.oper == '=':
            return SqlCompareOper('!=', self.left, self.right)
        elif self.oper == '!=':
            return SqlCompareOper('=', self.left, self.right)
        elif self.oper == 'LIKE':
            return SqlCompareOper('NOT LIKE', self.left, self.right)
        elif self.oper == 'NOT LIKE':
            return SqlCompareOper('LIKE', self.left, self.right)
        elif self.oper == 'BETWEEN':
            return SqlCompareOper('NOT BETWEEN', self.left, self.right)
        elif self.oper == 'NOT BETWEEN':
            return SqlCompareOper('BETWEEN', self.left, self.right)
        elif self.oper == 'IN':
            return SqlCompareOper('NOT IN', self.left, self.right)
        elif self.oper == 'NOT IN':
            return SqlCompareOper('IN', self.left, self.right)
        elif self.oper == '<':
            return SqlCompareOper('>=', self.left, self.right)
        elif self.oper == '<=':
            return SqlCompareOper('>', self.left, self.right)
        elif self.oper == '>':
            return SqlCompareOper('<=', self.left, self.right)
        elif self.oper == '>=':
            return SqlCompareOper('<', self.left, self.right)
        else:
            raise TypeError(f'{self.oper} not support NOT')


class SqlUnaryOper(SqlOper):
    __slots__ = 'oper', 'right'

    def __init__(self, oper: str, right: SqlExpr):
        super().__init__(oper)
        self.right = right

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self.oper)
        self.right.__sql_stat__(stat)

    def __repr__(self):
        return '(' + self.oper + repr(self.right) + ')'


class SqlCastOper(SqlOper):
    __slots__ = 'oper', 'right'

    def __init__(self, oper: str, right: SqlExpr):
        super().__init__(oper)
        self.right = right

    def __sql_stat__(self, stat: SqlStat):
        stat.add(['CAST', '('])
        self.right.__sql_stat__(stat)
        stat.add(['AS', self.oper, ')'])

    def __repr__(self):
        return 'CAST(' + repr(self.right) + ' AS ' + self.oper + ')'


class SqlOrderOper(SqlOper):
    __slots__ = 'oper', 'right'

    def __init__(self, oper: str, right: SqlExpr):
        super().__init__(oper)
        self.right = right

    def __sql_stat__(self, stat: SqlStat):
        self.right.__sql_stat__(stat)
        stat.add(self.oper)

    def nulls_first(self) -> Self:
        if self.oper in ('ASC', 'DESC'):
            return SqlOrderOper(f'{self.oper} NULLS FIRST', self.right)
        elif self.oper in ('NULLS FIRST', 'ASC NULLS FIRST', 'DESC NULLS FIRST'):
            return self
        else:
            raise RuntimeError(f'ORDER BY {self.oper} NULLS FIRST')

    def nulls_last(self) -> Self:
        if self.oper in ('ASC', 'DESC'):
            return SqlOrderOper(f'{self.oper} NULLS LAST', self.right)
        elif self.oper in ('NULLS LAST', 'ASC NULLS LAST', 'DESC NULLS LAST'):
            return self
        else:
            raise RuntimeError(f'ORDER BY {self.oper} NULLS LAST')

    def __repr__(self):
        return repr(self.right) + ' ' + self.oper


class SqlBinaryOper(SqlOper):
    __slots__ = 'oper', 'left', 'right'

    def __init__(self, oper: str, left: SqlExpr, right: Union[str, SqlExpr]):
        super().__init__(oper)
        self.left = left
        self.right = right

    def __sql_stat__(self, stat: SqlStat):
        self.left.__sql_stat__(stat)
        stat.add(self.oper)
        self.right.__sql_stat__(stat)

    def __repr__(self):
        return '(' + repr(self.left) + self.oper + repr(self.right) + ')'


class SqlVarArgOper(SqlOper):
    __slots__ = 'oper', 'args'

    def __init__(self, oper: str, args: list[SqlExpr]):
        super().__init__(oper)
        self.args = args

    def __sql_stat__(self, stat: SqlStat):
        if self.oper in ('NOT AND', 'NOT OR'):
            stat.add(['NOT', '('])
            self.not_().__sql_stat__(stat)
            stat.add(')')
        elif self.oper in ('AND', 'OR'):
            stat.add('(')
            sql_join(stat, self.oper, self.args)
            stat.add(')')
        else:
            raise ValueError()

    def __repr__(self):
        return self.oper + '(' + ','.join(map(repr, self.args)) + ')'

    def __invert__(self) -> SqlVarArgOper:
        return self.not_()

    def not_(self) -> SqlVarArgOper:
        if self.oper == 'AND':
            return SqlVarArgOper('NOT AND', self.args)
        elif self.oper == 'OR':
            return SqlVarArgOper('NOT OR', self.args)
        elif self.oper == 'NOT AND':
            return SqlVarArgOper('AND', self.args)
        elif self.oper == 'NOT OR':
            return SqlVarArgOper('OR', self.args)
        else:
            raise TypeError(f'NOT {self.oper}')

    def __and__(self, other: Union[str, SqlExpr]) -> SqlVarArgOper:
        if self.oper == 'AND':
            return SqlVarArgOper('AND', [*self.args, wrap(other)])
        return SqlVarArgOper('AND', [self, wrap(other)])

    def __or__(self, other: Union[str, SqlExpr]) -> SqlVarArgOper:
        if self.oper == 'OR':
            return SqlVarArgOper('OR', [*self.args, wrap(other)])
        return SqlVarArgOper('OR', [self, wrap(other)])


class SqlConcatOper(SqlOper):
    __slots__ = 'args',

    def __init__(self, args: list[SqlExpr], oper='||'):
        super().__init__(oper)
        self.oper = oper
        self.args = args

    def __sql_stat__(self, stat: SqlStat):
        sql_join(stat, self.oper, self.args)

    def __repr__(self):
        return '||'.join(map(repr, self.args))


class SqlFuncOper(SqlOper):
    __slots__ = 'oper', 'args'

    def __init__(self, oper: str, *args):
        super().__init__(oper)
        self.args = [wrap(it) for it in args]

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self.oper)

        if len(self.args):
            stat.add('(')
            sql_join(stat, ',', self.args)
            stat.add(')')
        else:
            stat.add('()')

    def __repr__(self):
        return self.oper + '(' + ','.join(map(repr, self.args)) + ')'


class SqlExistsOper(SqlOper):
    """https://www.sqlite.org/lang_expr.html#the_exists_operator"""
    __slots__ = 'oper', 'stat'

    def __init__(self, oper: Literal['EXISTS', 'NOT EXISTS'], stat: SqlStat):
        super().__init__(oper)
        self.stat = stat

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self.oper)
        stat.add(self.stat)

    def __invert__(self):
        if self.oper == 'EXISTS':
            return SqlExistsOper('NOT EXISTS', self.stat)
        elif self.oper == 'NOT EXISTS':
            return SqlExistsOper('EXISTS', self.stat)
        else:
            raise RuntimeError()


class SqlAggregateFunc(SqlFuncOper):

    def __init__(self, oper: str, *args):
        super().__init__(oper, *args)
        self._where: Optional[SqlVarArgOper] = None
        self._distinct = False

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self.oper)

        stat.add('(')
        if self._distinct:
            stat.add('DISTINCT')

        if len(self.args):
            sql_join(stat, ',', self.args)
        stat.add(')')

        if (where := self._where) is not None:  # type: SqlVarArgOper
            stat.add(['FILTER', '(', 'WHERE'])
            where.__sql_stat__(stat)
            stat.add(')')

    def distinct(self) -> Self:
        self._distinct = True
        return self

    def where(self, *args: Union[bool, SqlExpr]) -> Self:
        self._where = SqlVarArgOper('AND', list(map(wrap, args)))
        return self


class SqlWindowDef(SqlExpr):

    def __init__(self, name: str = None):
        self.name = name
        self._partition: Optional[list[SqlExpr]] = None
        self._order_by: Optional[list[SqlExpr]] = None
        self._frame_spec: SqlWindowFrame = None

    def __sql_stat__(self, stat: SqlStat):
        stat.add('(')
        if (p := self._partition) is not None:
            stat.add('PARTITION BY')
            sql_join(stat, ',', p)

        if (p := self._order_by) is not None:
            stat.add('ORDER BY')
            sql_join(stat, ',', p)

        if (spec := self._frame_spec) is not None:
            spec.__sql_stat__(stat)

        stat.add(')')

    def over(self, *, order_by=None, partition_by=None) -> Self:
        if isinstance(order_by, (tuple, list)):
            self._order_by = wrap_seq(*order_by)
        elif order_by is not None:
            self._order_by = [wrap(order_by)]

        if isinstance(partition_by, (tuple, list)):
            self._partition = wrap_seq(*partition_by)
        elif partition_by is not None:
            self._partition = [wrap(partition_by)]

        return self

    @overload
    def frame(self, on: Literal['RANGE', 'ROWS', 'GROUPS']) -> ContextManager[SqlWindowFrame]:
        pass

    @contextlib.contextmanager
    def frame(self, on):
        frame = SqlWindowFrame(on)
        yield frame
        self._frame_spec = frame


class SqlWindowFrame:
    class Spec:
        __slots__ = 'kind', 'expr'
        __match_args__ = 'kind', 'expr'

        def __init__(self, kind: Literal['UNBOUNDED PRECEDING', 'CURRENT ROW', 'PRECEDING', 'FOLLOWING'],
                     expr: SqlExpr = None):
            self.kind = kind
            self.expr = expr

        def __sql_stat__(self, stat: SqlStat):
            if self.kind in ('UNBOUNDED PRECEDING', 'CURRENT ROW'):
                stat.add(self.kind)
            elif self.kind in ('PRECEDING', 'FOLLOWING'):
                self.expr.__sql_stat__(stat)
                stat.add(self.kind)
            else:
                raise ValueError()

    def __init__(self, on: Literal['RANGE', 'ROWS', 'GROUPS']):
        self._on = on
        self._specs = []
        self._exclude = None

    def __sql_stat__(self, stat: SqlStat):
        stat.add(self._on)

        spec = self._specs[-1]
        if isinstance(spec, SqlWindowFrame.Spec) and spec.kind in ('UNBOUNDED PRECEDING', 'CURRENT ROW', 'PRECEDING'):
            spec.__sql_stat__(stat)
        elif isinstance(spec, tuple) and len(spec) == 2 and isinstance(spec[0], SqlWindowFrame.Spec) and isinstance(
                spec[1], SqlWindowFrame.Spec):
            stat.add('BETWEEN')
            spec[0].__sql_stat__(stat)
            stat.add('AND')
            spec[1].__sql_stat__(stat)
        else:
            raise TypeError()

        if (ex := self._exclude) is not None:
            stat.add(['EXCLUDE', ex])

    # noinspection PyUnusedLocal
    def between(self, left, right) -> Self:
        *_, left, right = self._specs
        self._specs.append((left, right))
        return self

    def unbounded_preceding(self) -> Self:
        ret = self.Spec('UNBOUNDED PRECEDING')
        self._specs.append(ret)
        return self

    def current_row(self) -> Self:
        ret = self.Spec('CURRENT ROW')
        self._specs.append(ret)
        return self

    def preceding(self, expr) -> Self:
        ret = self.Spec('PRECEDING', wrap(expr))
        self._specs.append(ret)
        return self

    def following(self, expr) -> Self:
        ret = self.Spec('FOLLOWING', wrap(expr))
        self._specs.append(ret)
        return self

    def exclude(self, on: Literal['NO OTHERS', 'CURRENT ROW', 'GROUP', 'TIES']):
        self._exclude = on


class SqlWindowFunc(SqlAggregateFunc):

    def __init__(self, oper: str, *args):
        SqlAggregateFunc.__init__(self, oper, *args)
        self._over: Union[str, SqlWindowDef, None] = None

    def __sql_stat__(self, stat: SqlStat):
        super().__sql_stat__(stat)

        if self._over is not None:
            stat.add('OVER')
            if isinstance(self._over, str):
                stat.add(self._over)
            elif isinstance(self._over, SqlWindowDef):
                if self._over.name is None:
                    self._over.__sql_stat__(stat)
                elif isinstance(self._over.name, str):
                    stat.add(self._over.name)
                    stat.window_def(self._over)
                else:
                    raise TypeError()
            elif isinstance(self._over, SqlAlias) and isinstance(self._over._value, SqlWindowDef):
                stat.add(self._over._name)
                stat.window_def(self._over._value)
            else:
                raise TypeError()

    @overload
    def over(self, name: Union[str, SqlWindowDef, SqlAlias[SqlWindowDef]]) -> Self:
        pass

    @overload
    def over(self, *, order_by=None, partition_by=None) -> Self:
        pass

    def over(self, name: str = None, *, order_by=None, partition_by=None) -> Self:
        if name is None:
            self._over = over = SqlWindowDef()
            over.over(order_by=order_by, partition_by=partition_by)
        elif isinstance(name, str):
            self._over = name
        elif isinstance(name, SqlWindowDef):
            self._over = name
        elif isinstance(name, SqlAlias) and isinstance(name._value, SqlWindowDef):
            self._over = name
        else:
            raise TypeError()

        return self


class SqlCaseExpr(SqlExpr):
    """https://www.sqlite.org/lang_expr.html#the_case_expression"""

    __slots__ = 'expr', 'cases'
    __match_args__ = 'expr', 'cases'

    def __init__(self, expr: SqlExpr = None):
        self.expr: Optional[SqlExpr] = expr
        self.cases: list[tuple[Optional[SqlExpr], SqlExpr]] = []

    def __sql_stat__(self, stat: SqlStat):
        stat.add('CASE')
        if self.expr is not None:
            self.expr.__sql_stat__(stat)

        for case, then in self.cases:  # type: SqlExpr , SqlExpr
            if case is not None:
                stat.add('WHEN')
                case.__sql_stat__(stat)
                stat.add('THEN')
            else:
                stat.add('ELSE')
            then.__sql_stat__(stat)

        stat.add('END')

    def when(self, case, then) -> Self:
        if len(self.cases) and self.cases[-1][0] is None:
            raise RuntimeError('when after else')

        if isinstance(case, str):
            case = SqlLiteral(repr(case))
        else:
            case = wrap(case)

        if isinstance(then, str):
            then = SqlLiteral(repr(then))
        else:
            then = wrap(then)

        self.cases.append((case, then))
        return self

    def else_(self, then) -> Self:
        if len(self.cases) == 0:
            raise RuntimeError('else without when')
        elif len(self.cases) and self.cases[-1][0] is None:
            raise RuntimeError('else after else')

        if isinstance(then, str):
            then = SqlLiteral(repr(then))
        else:
            then = wrap(then)

        self.cases.append((None, then))
        return self


class SqlCteExpr(SqlExpr):
    __slots__ = '_name', '_select'

    def __init__(self, name: str, select: SqlSelectStat):
        self._name = name
        self._select = select
        select._connection = None

    def __getattr__(self, item) -> SqlAliasField:
        return SqlAliasField(self, self._name, item)

    def __sql_stat__(self, stat: SqlStat):
        stat.add(['WITH', self._name, 'AS'])
        stat.add(self._select)



"""
Suite2p Result Parser & Utils
===============================

:author:
    Yu-Ting Wei

This module provide the container for suite2p registration/segmentation 2P results.

Also include utilities for signal processing and basic plotting

.. code-block:: python

    s2p_dir = ...  # suite2p base directory (*/suite2p/plane*)
    s2p = Suite2PResult.load(s2p_dir, cell_prob=0.0, channel=0)

See available attributes/properties/methods
--------------------------------------------
.. code-block:: python

    print(dir(s2p))


See the option config for running the GUI
---------------------------------------------

.. code-block:: python

    from rich.pretty import pprint
    pprint(s2p.ops)

"""
from .core import *
from .plot import *
from .signals import *



import subprocess
from pathlib import Path
from typing import Literal

import torch.cuda

from neuralib.argp import argument, as_argument
from neuralib.segmentation.base import AbstractSegmentationOption
from neuralib.segmentation.cellpose.core import AbstractCellPoseOption, CPOSE_MODEL
from neuralib.util.cli_args import CliArgs
from neuralib.util.gpu import check_mps_available
from neuralib.util.util_verbose import fprint

__all__ = ['CellPoseSubprocOption']


class CellPoseSubprocOption(AbstractCellPoseOption):
    DESCRIPTION = 'Run cellpose as a CLI subprocess. Detail refer to Cellpose.cli'

    file: Path = as_argument(AbstractSegmentationOption.file).with_options(
        metavar='FILE/DIR',
        required=True,
        help='run one single image if file or folder',
    )

    model: CPOSE_MODEL = as_argument(AbstractSegmentationOption.model).with_options(default='cyto3')

    label_type: Literal['gcamp6s', 'retro-aav'] = argument(
        '-T', '--type',
        default='retro-aav',
        help='type of fluorescence label type'
    )

    no_norm: bool = argument(
        '--no-norm',
        help='do not normalize images (normalize=False)'
    )

    def run(self):
        # single
        if self.file.is_file():

            if self.seg_result.exists():  # view
                if self.napari_view:
                    self.launch_napari()
                elif self.cellpose_gui:
                    self.launch_cellpose_gui()
                else:
                    raise RuntimeError(f'result for {self.file} existed!')
            else:
                self.eval()

        # batch
        else:
            self.eval()

    def eval(self):
        cmds = self.build_cli()
        code = subprocess.check_call(cmds)
        return code

    def build_cli(self) -> list[str]:
        ret = ['python', '-m', 'cellpose', '--verbose']

        #
        if self.file.is_dir():
            ret.extend(CliArgs('--dir', str(self.file)).as_command())
        elif self.file.is_file():
            ret.extend(CliArgs('--image_path', str(self.file)).as_command())

        #
        ret.extend(CliArgs('--chan', self.chan_seg).as_command())
        ret.extend(CliArgs('--chan2', self.chan_nuclear).as_command())

        #
        ret.extend(CliArgs('--pretrained_model', self.model).as_command())

        #
        ret.extend(CliArgs('--diameter', str(self.diameter)).as_command())

        #
        if self.no_norm:
            ret.extend(CliArgs('--no_norm').as_command())

        #
        if torch.cuda.is_available() or check_mps_available():
            ret.extend(CliArgs('--use_gpu').as_command())

            if check_mps_available():
                ret.extend(CliArgs('--gpu_device', 'mps').as_command())

        fprint(ret)

        return ret


if __name__ == '__main__':
    CellPoseSubprocOption().main()



from __future__ import annotations

from typing import Literal, get_args, Iterable

import polars as pl

from neuralib.util.util_verbose import fprint

__all__ = [
    'ROIS_NORM_TYPE',
    'MouseBrainRoiNormHandler',
    'handle_failure_norm',
    'foreach_norm_method'
]

ROIS_NORM_TYPE = Literal['channel', 'volume', 'cell']


class MouseBrainRoiNormHandler:
    """Class for handle the normalization for ROIs across mouse brain"""

    def __init__(self, norm_type: ROIS_NORM_TYPE | None = None):
        """
        :param norm_type: which kind of normalization for the roi labeling

            ``channel``: normalize to fraction of rois for a specific color fluorescence channel

            ``volume``: normalize to the volume size per region (cellatlas-based) # TODO validate with allenSDK source

            ``cell``: normalize to the total cell counts per region (cellatlas-based)
        """
        if norm_type is not None and norm_type not in get_args(ROIS_NORM_TYPE):
            raise ValueError('')

        self.norm_type = norm_type

    @classmethod
    def handle_failure(cls, df: pl.DataFrame, *,
                       expr: pl.Expr | None = None,
                       drop_not_found: bool = True) -> pl.DataFrame:
        return handle_failure_norm(df, expr=expr, drop_not_found=drop_not_found)

    @property
    def unit_label(self) -> str:
        """For plotting"""
        if self.norm_type is None:
            return 'Counts#'
        elif self.norm_type == 'cell':
            return 'Cell density(%)'
        elif self.norm_type == 'volume':
            return 'Density(cells-mm^3)'
        elif self.norm_type == 'channel':
            return 'Fraction of input(%)'
        else:
            raise ValueError('')

    @property
    def unit_col(self) -> str:
        """For dataframe column"""
        if self.norm_type is None:
            return 'n_rois'
        elif self.norm_type == 'channel':
            return 'percent'
        elif self.norm_type in ('volume', 'cell'):
            return f'{self.norm_type}_norm_n_rois'
        else:
            raise ValueError('')

    def expand(self,
               df: pl.DataFrame,
               area_col: str,
               expand_cols: list[str]) -> pl.DataFrame:
        """
        expand the dataframe based on different normalization

        :param df: input dataframe has abs cell counts per region
        :param area_col: brain region acronym for querying
        :param expand_cols: columns with abs cell counts need to be expanded
        :return: expanded dataframe
        """
        if self.norm_type in ('cell', 'volume'):
            ret = self._compute_cellatlas_expand_field(df, area_col, expand_cols)
        elif self.norm_type == 'channel':  # calculate in RoiClassifier.percent_sorted_df
            ret = df
        else:
            raise ValueError(f'{self.norm_type} unknown')

        return ret

    def _compute_cellatlas_expand_field(self, df, area_col, expand_cols) -> pl.DataFrame:
        from neuralib.atlas.cellatlas import CellAtlas
        ctlas = CellAtlas.load_sync_allen_structure_tree()

        if self.norm_type == 'cell':
            calc = ((pl.col(c) / pl.col('n_neurons') * 100).alias(f'cell_norm_{c}') for c in expand_cols)
        elif self.norm_type == 'volume':
            calc = ((pl.col(c) / pl.col('Volumes [mm^3]')).alias(f'volume_norm_{c}') for c in expand_cols)

        sync = (
            df.join(ctlas.rename({'acronym': area_col}), on=area_col, how='left')
            .with_columns(calc)
            .fill_null(-1)
            .fill_nan(-1)
        )

        return sync


def handle_failure_norm(df: pl.DataFrame, *,
                        expr: pl.Expr | None = None,
                        drop_not_found: bool = True):
    """
    warning verbose or drop if
    cellatlas cannot find the n_neurons OR volume in the given acronym,

    :param df:
    :param expr: condition for determine if it's failure
    :param drop_not_found: whether drop the region that can not be found
    """
    if expr is None:
        expr = pl.col('name').is_null()

    failure = df.filter(expr)
    if failure.shape[0] == 0:
        return df

    msg = '\n#===== WARNING =====#'
    msg += f'\n{failure}'
    msg += '\nIF IMPORTANT INFO LOSS, please decrease the merge level ' \
           'or check details in cell atlas APP/csv & allenccf raw csv!!!'

    if drop_not_found:
        msg += '\n#===== DROP AREAS ======#'
        df = df.filter(~expr)

    fprint(msg, vtype='warning')

    return df


def foreach_norm_method(include_abs_count: bool = True, **kwargs) -> Iterable[MouseBrainRoiNormHandler]:
    """
    foreach normalization method

    :param include_abs_count: if iterator include abs cell counts
    :param kwargs: pass to ``MouseBrainRoiNormHandler``
    :return:
    """
    methods = []
    if include_abs_count:
        methods = [None]
    methods += get_args(ROIS_NORM_TYPE)

    for t in methods:
        yield MouseBrainRoiNormHandler(t, **kwargs)



from __future__ import annotations

import sqlite3
import warnings
from collections.abc import Iterator
from typing import overload, TYPE_CHECKING, Any, TypeVar, Generic, Optional, Literal, Union, cast, Callable

import polars as pl
from typing_extensions import Self

from .expr import *
from .expr import sql_join_set
from .table import *

if TYPE_CHECKING:
    from .connection import Connection

__all__ = [
    'SqlStat',
    'SqlSelectStat',
    'SqlInsertStat',
    'SqlUpdateStat',
    'SqlDeleteStat',
    'Cursor',
]

T = TypeVar('T')
S = TypeVar('S')


class SqlStat(Generic[T]):
    """Abstract SQL statement."""

    def __init__(self, table: Optional[type[T]]):
        self.table = table
        self._stat: list[str] = []
        self._para: list[Any] = []
        self._deparameter = False

        from .connection import get_connection_context
        self._connection: Optional[Connection] = get_connection_context()
        # connection will be set to None when following situation happen:
        # * after submit()
        # * __del__() call submit()
        # * as a subquery be used by other statement.
        # TODO ...

    def build(self) -> str:
        """build a SQL statement."""
        ret = ' '.join(self._stat)
        self._stat = None
        return ret

    def __str__(self) -> str:
        table = table_name(self.table) if self.table is not None else '...'
        return f'{type(Self).__name__}[{table}]'

    def __repr__(self) -> str:
        ret = []
        par = list(self._para)
        for value in self._stat:
            if value == '?':
                ret.append(f'?({repr(par.pop(0))})')
            else:
                ret.append(value)
        return ' '.join(ret)

    def submit(self) -> Cursor[T]:
        """
        build the SQL statement and execute.

        :return: a cursor
        :raise RuntimeError: current statement does not bind with a connection.
        """
        if (connection := self._connection) is None:
            raise RuntimeError('Do not in a connection context')

        ret = Cursor(connection, connection.execute(self), self.table)
        self._connection = None
        return ret

    def __del__(self):
        # check connection and auto_commit,
        # make SqlStat auto submit itself when nobody is referring to it,
        # so users do need to explict call submit() for every statements.
        if self._connection is not None and self._stat is not None and len(self._stat) > 0:
            try:
                self.submit()
            except BaseException as e:
                warnings.warn(repr(e))

    @overload
    def add(self, stat: SqlStat) -> Self:
        pass

    @overload
    def add(self, stat: Union[str, list[str]], *para) -> Self:
        pass

    def add(self, stat, *para) -> Self:
        """
        Add SQL token.
        """
        if self._stat is None:
            raise RuntimeError("Statement is closed.")

        if isinstance(stat, SqlStat):
            self._stat.append('(')
            self._stat.extend(stat._stat)
            self._stat.append(')')
            self._para.extend(stat._para)
            stat._connection = None
        elif isinstance(stat, (tuple, list)):
            if self._deparameter:
                para = list(para)
                for _stat in stat:
                    if _stat == '?':
                        self._stat.append(repr(para.pop(0)))
                    else:
                        self._stat.append(_stat)
            else:
                self._stat.extend(stat)
                self._para.extend(para)
        else:
            if self._deparameter:
                if stat == '?':
                    self._stat.append(repr(para[0]))
                else:
                    self._stat.append(stat)
            else:
                self._stat.append(stat)
                self._para.extend(para)

        return self


class Cursor(Generic[T]):
    """
    A SQL cursor wrapper.

    It will try to cast to T from the tuple returns.

    """

    def __init__(self, connection: Connection, cursor: sqlite3.Cursor, table: type[T] = None):
        self._connection = connection
        self._cursor = cursor
        self._table = table

        header = cursor.description
        self.headers: list[str] = [it[0] for it in header] if header is not None else []
        """headers of return"""

        self._table_cls = None
        if table is not None:
            from .table import table_class
            try:
                self._table_cls = table_class(table)
            except AttributeError:
                pass

    def __del__(self):
        self._cursor.close()
        self._connection = None

    def fetchall(self) -> list[T]:
        """fetch all results."""
        return list(self)

    def fetchone(self) -> Optional[T]:
        """fetch the first result."""
        if (ret := self._cursor.fetchone()) is None:
            return None

        if self._table_cls is not None:
            return self._table_cls.table_new(*ret)

        return ret

    def __iter__(self) -> Iterator[T]:
        """iterate the results."""
        table_cls = self._table_cls
        for a in self._cursor:
            if table_cls is not None:
                yield table_cls.table_new(*a)
            else:
                yield a

    def fetch_polars(self) -> pl.DataFrame:
        return pl.DataFrame(list(self._cursor), schema=self.headers)


class SqlWhereStat:
    """statement with **WHERE** support."""

    def where(self, *expr: Union[bool, SqlCompareOper, SqlExpr, Any, None]) -> Self:
        """
        ``WHERE`` clause: https://www.sqlite.org/lang_select.html#whereclause

        >>> select_from(A).where( # doctest: SKIP
        ...     A.a == 1, A.b == 2
        ... ).build()
        SELECT * FROM A
        WHERE (A.a = 1) AND (A.b = 2)

        :param expr:
        :return:
        """
        zelf = cast(SqlStat, self)
        expr = [it for it in expr if it is not None]
        if len(expr):
            from .func_stat import and_
            zelf.add('WHERE')
            and_(*expr).__sql_stat__(zelf)
        return self


class SqlLimitStat:
    @overload
    def limit(self, n: int) -> Self:
        pass

    @overload
    def limit(self, row_count: int, offset: int) -> Self:
        pass

    def limit(self, *args: int) -> Self:
        """
        ``LIMIT``: https://www.sqlite.org/lang_select.html#limitoffset

        >>> select_from(A).limit(10).build() # doctest: SKIP
        SELECT * FROM A LIMIT 10
        >>> select_from(A).limit(10, 10).build() # doctest: SKIP
        SELECT * FROM A LIMIT 10 OFFSET 10

        **NOTE**

        LIMIT on UPDATE/DELETE need compile flag ``SQLITE_ENABLE_UPDATE_DELETE_LIMIT``.

        >>> assert Connection().sqlite_compileoption_used('SQLITE_ENABLE_UPDATE_DELETE_LIMIT') # doctest: SKIP

        """
        zelf = cast(SqlStat, self)

        if len(args) == 1:
            n, *_ = args
            zelf.add(['LIMIT', str(n)])
        elif len(args) == 2:
            row_count, offset = args
            zelf.add(['LIMIT', str(row_count), 'OFFSET', str(offset)])
        else:
            raise TypeError()
        return self

    def order_by(self, *by: Union[int, str, SqlExpr, Any]) -> Self:
        """
        ``ORDER BY``: https://www.sqlite.org/lang_select.html#orderby

        >>> select_from(A).order_by(A.a).build() # doctest: SKIP
        SELECT * FROM A ORDER BY A.

        **possible ordering**

        >>> select_from(A).order_by( # doctest: SKIP
        ...     asc(A.a), desc(A.b), nulls_first(A.c), asc(A.d).nulls_last(),
        ... )
        SELECT * FROM A ORDER BY
        A.a ASC, A.b DESC, A.c NULLS FIRST, A.d ASC NULLS LAST

        **NOTE**

        ORDER BY on UPDATE/DELETE need compile flag ``SQLITE_ENABLE_UPDATE_DELETE_LIMIT``.

        >>> assert Connection().sqlite_compileoption_used('SQLITE_ENABLE_UPDATE_DELETE_LIMIT') # doctest: SKIP

        """
        zelf = cast(SqlStat, self)
        zelf.add('ORDER BY')
        fields = []
        for it in by:
            if isinstance(it, int):
                fields.append(SqlLiteral(str(it)))
            elif isinstance(it, str):
                fields.append(SqlLiteral(it))
            elif isinstance(it, SqlExpr):
                fields.append(it)
            else:
                fields.append(wrap(it))

        SqlConcatOper(fields, ',').__sql_stat__(zelf)

        return self


class SqlSelectStat(SqlStat[T], SqlWhereStat, SqlLimitStat, Generic[T]):
    """**SELECT** statement."""

    def __init__(self, table: Optional[type[T]]):
        super().__init__(table)
        self._involved: list[Union[type, SqlAlias[type]]] = []
        self._window_defs: dict[str, SqlWindowDef] = {}

    def __matmul__(self, other: str) -> SqlAlias[SqlSubQuery]:
        """wrap itself as a subquery with an alias name."""
        return SqlAlias(SqlSubQuery(self), other)

    def fetchall(self) -> list[T]:
        """submit and fetch all result."""
        return self.submit().fetchall()

    def fetchone(self) -> Optional[T]:
        """submit and fetch the first result."""
        return self.submit().fetchone()

    def __iter__(self) -> Iterator[T]:
        """submit and iterate the results"""
        return iter(self.submit())

    def fetch_polars(self) -> pl.DataFrame:
        """submit and fetch all as a polar dataframe."""
        return self.submit().fetch_polars()

    def window_def(self, *windows: Union[SqlWindowDef, SqlAlias[SqlWindowDef]], **window_ks: SqlWindowDef) -> Self:
        """
        define windows.

        """
        if (defs := self._window_defs) is None:
            raise RuntimeError('not a select statement')

        for window in windows:
            if isinstance(window, SqlWindowDef):
                if window.name is None:
                    raise RuntimeError('?? AS ' + repr(window))
                defs[window.name] = window
            elif isinstance(window, SqlAlias) and isinstance(window._value, SqlWindowDef):
                defs[window.name] = window._value
            else:
                raise TypeError()

        defs.update(window_ks)
        return self

    @overload
    def join(self, table: Union[type[S], SqlSelectStat[S], SqlAlias[S]], *,
             by: Literal['left', 'right', 'inner', 'full outer'] = None) -> SqlJoinStat:
        pass

    @overload
    def join(self, table: Union[type[S], SqlSelectStat[S], SqlAlias[S]], *,
             by: Literal['cross']) -> SqlSelectStat[tuple]:
        pass

    def join(self, table: Union[type[S], SqlSelectStat[S], SqlAlias[S]], *,
             by: Literal['left', 'right', 'inner', 'full outer', 'cross'] = None):
        """
        ``JOIN`` https://www.sqlite.org/lang_select.html#strange_join_names

        >>> select_from(A.a, B.b).join(B).on(A.a == B.a) # doctest: SKIP
        SELECT A.a, B.b FROM A
        JOIN B ON A.a = B.a

        """
        if by is not None:
            self.add(by.upper())
        self.add('JOIN')

        if isinstance(table, type):
            self.add(table_name(table))
            that = [table]
        elif isinstance(table, SqlSelectStat):
            self.add(table)
            that = []
        elif isinstance(table, SqlAlias) and isinstance(that_table := table._value, type):
            self.add(table_name(that_table))
            self.add(table._name)
            that = [table]
        elif isinstance(table, SqlAlias) and isinstance(table._value, SqlSubQuery) and isinstance(table._value.stat, SqlSelectStat):
            self.add(table._value.stat)
            self.add('AS')
            self.add(table._name)
            that = []
        else:
            raise TypeError(f'JOIN {table}')

        self.table = None

        if by == 'cross':
            self._involved.extend(that)
            return self

        return SqlJoinStat(self, that)

    def group_by(self, *by) -> Self:
        """
        ``GROUP BY`` https://www.sqlite.org/lang_select.html#resultset
        """
        if len(by) == 0:
            raise RuntimeError()

        self.add('GROUP BY')
        for field in by:
            if isinstance(field, (int, float, bool)):
                self.add(str(field))
            elif isinstance(field, str):
                self.add(repr(field))
            elif isinstance(field, Field):
                self.add(f'{field.table_name}.{field.name}')
            elif isinstance(field, SqlField):
                self.add(f'{field.table_name}.{field.name}')
            elif isinstance(field, SqlAlias):
                self.add(field._name)
            elif isinstance(field, SqlExpr):
                field.__sql_stat__(self)
            else:
                raise TypeError('GROUP BY ' + repr(field))
            self.add(',')

        self._stat.pop()
        return self

    def having(self, *exprs: Union[bool, SqlExpr]) -> Self:
        """
        ``HAVING`` https://www.sqlite.org/lang_select.html#resultset
        """
        if len(exprs) == 0:
            return self

        from .func_stat import and_
        self.add('HAVING')
        and_(*exprs).__sql_stat__(self)

        return self

    def intersect(self, stat: SqlStat) -> Self:
        """
        ``INTERSECT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        self.add('INTERSECT')
        self.add(stat._stat, *stat._para)
        stat._connection = None
        return self

    def __and__(self, other: SqlStat) -> Self:
        """
        ``INTERSECT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        return self.intersect(other)

    def union(self, stat: SqlStat, all=False) -> Self:
        """
        ``UNION`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        self.add('UNION')
        if all:
            self.add('ALL')
        self.add(stat._stat, *stat._para)
        stat._connection = None
        return self

    def __or__(self, other: SqlStat) -> Self:
        """
        ``UNION`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        return self.union(other)

    def except_(self, stat: SqlStat) -> Self:
        """
        ``EXCEPT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        self.add('EXCEPT')
        self.add(stat._stat, *stat._para)
        stat._connection = None
        return self

    def __sub__(self, other: SqlStat) -> Self:
        """
        ``EXCEPT`` https://www.sqlite.org/lang_select.html#compound_select_statements
        """
        return self.except_(other)


class SqlJoinStat:
    def __init__(self, stat: SqlSelectStat[tuple], join: list[Union[type, SqlAlias[type]]]):
        self._stat = stat
        self._this = stat._involved
        self._that = join

    def using(self, *fields) -> SqlSelectStat[tuple]:
        if len(fields) == 0:
            raise RuntimeError()

        stat = self._stat
        stat.add(['USING', '('])

        for field in fields:
            if isinstance(field, str):
                stat.add(repr(field))
            elif isinstance(field, Field):
                stat.add(field.name)
            elif isinstance(field, SqlField):
                stat.add(field.name)
            else:
                raise TypeError('USING ' + repr(field))
            stat.add(',')

        stat._stat.pop()
        stat.add(')')

        stat._involved.extend(self._that)
        return stat

    def on(self, *exprs: Union[bool, SqlExpr]) -> SqlSelectStat[tuple]:
        if len(exprs) == 0:
            raise RuntimeError()

        self._stat.add('ON')
        from .func_stat import and_
        and_(*exprs).__sql_stat__(self._stat)

        self._stat._involved.extend(self._that)
        return self._stat

    def by(self, constraint: Union[Callable, ForeignConstraint]) -> SqlSelectStat[tuple]:
        if not isinstance(constraint, ForeignConstraint):
            if (constraint := table_foreign_field(constraint)) is None:
                raise RuntimeError('not a foreign constraint')

        stat = self._stat

        if constraint.fields == constraint.foreign_fields:
            stat.add(['USING', '('])

            for field in constraint.fields:
                stat.add(field)
                stat.add(',')

            stat._stat.pop()
            stat.add(')')

        else:
            this = that = None
            for _this in self._this:
                if isinstance(_this, type):
                    if _this == constraint.table:
                        this = _this.__name__
                        break
                    elif _this == constraint.foreign_table:
                        that = _this.__name__
                        break

                if isinstance(_this, SqlAlias) and (_type := _this._value, type):
                    if _type == constraint.table:
                        this = _this._name
                        break
                    elif _type == constraint.foreign_table:
                        that = _this._name
                        break

                raise TypeError('internal error')
            else:
                raise RuntimeError('improper foreign constraint')

            for _that in self._that:
                if isinstance(_that, type):
                    if this is None and _that == constraint.table:
                        this = _this.__name__
                        break
                    if that is None and _that == constraint.foreign_table:
                        that = _this.__name__
                        break
                if isinstance(_that, SqlAlias) and isinstance(_type := _that._value, type):
                    if this is None and _type == constraint.table:
                        this = _that._name
                        break
                    if that is None and _type == constraint.foreign_table:
                        that = _that._name
                        break
                raise TypeError('internal error')
            else:
                raise RuntimeError('improper foreign constraint')

            assert this is not None and that is not None

            stat.add(['ON', '('])
            for af, bf in zip(constraint.fields, constraint.foreign_fields):
                stat.add([f'{this}.{af} = {that}.{bf}'])
                stat.add('AND')

            stat._stat.pop()
            stat.add(')')

        stat._involved.extend(self._that)
        return stat


class SqlReturnStat:
    def returning(self, *expr: Union[str, Any]) -> Self:
        zelf = cast(SqlStat, self)
        zelf.add('RETURNING')
        if len(expr) == 0:
            zelf.add('*')
        else:
            fields = []
            for exp in expr:
                if isinstance(exp, str):
                    zelf.add(exp)
                    fields.append(exp)
                elif isinstance(exp, SqlField):
                    zelf.add(exp.name)
                    fields.append(exp.name)
                elif isinstance(exp, SqlAlias) and isinstance(field := exp._value, SqlField):
                    zelf.add([field.name, 'AS', exp._name])
                    fields.append(exp._name)
                elif isinstance(exp, SqlAlias) and isinstance(_expr := exp._value, SqlExpr):
                    zelf._deparameter = True
                    _expr.__sql_stat__(zelf)
                    zelf._deparameter = False

                    zelf.add(['AS', exp._name])
                    fields.append(exp._name)
                else:
                    raise TypeError(f'RETURNING ({exp})')

                zelf.add(',')
            zelf._stat.pop()

            if isinstance(self, SqlInsertStat):
                self.table = None
                self._fields = fields

        return self


class SqlInsertStat(SqlStat[T], SqlReturnStat, Generic[T]):
    def __init__(self, table: type[T], fields: list[str] = None, *, named: bool = False):
        super().__init__(table)
        self._fields = fields

        # when
        #   None: `VALUES` set
        #   'DEFAULT': `DEFAULT VALUES` unset
        #   {field->value}: `DEFAULT VALUES` set
        self._values: Union[Literal['DEFAULT'], dict[str, Any], None] = {
            name: SqlLiteral(f':{name}') if named else SqlLiteral('?')
            for name in table_field_names(table)
        }
        self._named = named
        self._returning = False

    @overload
    def select_from(self, table: type[T], *, distinct: bool = False) -> SqlSelectStat[T]:
        pass

    @overload
    def select_from(self, *field, distinct: bool = False,
                    from_table: Union[str, type, SqlAlias, SqlSelectStat] = None) -> SqlSelectStat[tuple]:
        pass

    def select_from(self, *args, distinct: bool = False,
                    from_table: Union[str, type, SqlAlias, SqlSelectStat] = None) -> SqlSelectStat[T]:
        if isinstance(self._values, str):
            raise RuntimeError()

        self._connection = None

        self.add('(')
        for i, field in enumerate(self._values):
            if i > 0:
                self.add(',')
            self.add(field)
        self.add(')')

        from .stat_start import select_from
        ret = select_from(*args, distinct=distinct, from_table=from_table)
        ret._stat = [*self._stat, *ret._stat]

        return ret

    def values(self, *args, **kwargs: Union[str, SqlExpr]) -> Self:
        # TODO support direct set values
        if self._values is None or isinstance(self._values, str):
            raise RuntimeError()

        for expr in args:
            if isinstance(expr, SqlCompareOper) and expr.oper == '=' and isinstance(expr.left, SqlField):
                self._values[expr.left.field.name] = expr.right
            else:
                raise TypeError(expr)

        for field, expr in kwargs.items():
            if field not in self._values:
                raise RuntimeError(f'{table_name(self.table)}.{field} not found')
            self._values[field] = expr
        return self

    def defaults(self) -> Self:
        """insert default values"""
        self._values = 'DEFAULT'
        return self

    def _set_values(self):
        if isinstance(self._values, str):
            self.add(['DEFAULT', 'VALUES'])
        elif isinstance(self._values, dict):
            if self._fields is None:
                fields = list(self._values.keys())
            else:
                fields = self._fields

            self.add(['VALUES', '('])

            for field in fields:
                value = self._values[field]
                if isinstance(value, (int, float, bool, str)):
                    self.add(repr(value))
                elif isinstance(value, SqlLiteral):
                    self.add(value.value)
                elif isinstance(value, SqlPlaceHolder):
                    self.add(repr(value.value))
                elif isinstance(value, SqlStat):
                    self.add(value)
                elif isinstance(value, SqlExpr):
                    self._deparameter = True
                    value.__sql_stat__(self)
                    self._deparameter = False
                else:
                    raise TypeError(repr(value))
                self.add(',')
            self._stat.pop()
            self.add(')')

        self._values = None

    def on_conflict(self, *conflict, where: Union[bool, SqlCompareOper] = None) -> SqlUpsertStat[T]:
        self._set_values()
        return SqlUpsertStat(self, *conflict, where=where)

    def returning(self, *expr: Union[str, SqlExpr]) -> SqlInsertStat[tuple]:
        self._set_values()
        super().returning(*expr)
        self._returning = True
        return self

    def build(self) -> str:
        self._set_values()
        return super().build()

    def submit(self, parameter: list[T] = ()) -> Cursor[T]:
        if (connection := self._connection) is None:
            raise RuntimeError('Do not in a connection context')

        from .table import table_class
        try:
            table_cls = table_class(self.table)
        except AttributeError:
            pass
        else:
            if self._named:
                def mapper(p):
                    if isinstance(p, self.table):
                        return table_cls.table_dict(p)
                    return dict(p)
            else:
                def mapper(p):
                    if isinstance(p, self.table):
                        return table_cls.table_seq(p)
                    return tuple(p)

            parameter = list(map(mapper, parameter))

        if len(parameter):
            if self._returning:
                if len(parameter) > 1:
                    raise RuntimeError('only support return one data')
                cur = connection.execute(self, parameter[0])
            else:
                cur = connection.execute_batch(self, parameter)
        else:
            cur = connection.execute(self, parameter)

        ret = Cursor(connection, cur, self.table)
        self._connection = None
        return ret


class SqlUpsertStat(Generic[T]):
    """

    https://www.sqlite.org/syntax/upsert-clause.html

    """

    def __init__(self, stat: SqlInsertStat[T], *conflict, where: Union[bool, SqlCompareOper] = None):
        self._stat = stat
        self._conflict = conflict
        self._stat.add(['ON', 'CONFLICT'])
        self._do_where = False

        if len(conflict):
            from .stat_start import select_from_fields
            table, fields = select_from_fields(*conflict)
            if table is None:
                table = stat.table
            elif table != stat.table:
                raise RuntimeError()

            self._stat.add('(')
            for i, field in enumerate(fields):
                if i > 0:
                    self._stat.add(',')

                if isinstance(field, SqlField):
                    if field.table != table:
                        raise RuntimeError(f'field {field.table_name}.{field.name} not belong to {table.__name__}')
                    self._stat.add(f'{field.name}')
                elif isinstance(field, SqlLiteral) and isinstance(field.value, str):
                    self._stat.add(field.value)
                else:
                    raise RuntimeError(f'ON CONFLICT ({field}:{type(field).__name__})')

            if where is not None:
                self._stat.add('WHERE')
                self._stat._deparameter = True
                where.__sql_stat__(self._stat)
                self._stat._deparameter = False
            self._stat.add(')')

    def do_nothing(self) -> SqlInsertStat[T]:
        self._stat.add(['DO', 'NOTHING'])
        return self._stat

    def do_update(self, *args: Union[bool, SqlCompareOper], where: Union[bool, SqlCompareOper] = None) -> SqlInsertStat[T]:
        self._stat.add(['DO', 'UPDATE', 'SET'])

        self._stat._deparameter = True  # TODO
        sql_join_set(self._stat, ',', args)
        self._stat._deparameter = False

        if where is not None:
            self._stat.add('WHERE')
            self._stat._deparameter = True
            where.__sql_stat__(self._stat)
            self._stat._deparameter = False
        return self._stat


class SqlUpdateStat(SqlStat[T], SqlWhereStat, SqlLimitStat, SqlReturnStat, Generic[T]):
    def from_(self, query: Union[SqlStat, SqlAlias[SqlSubQuery]]) -> Self:
        self.add('FROM')
        if isinstance(query, SqlStat):
            self.add(query)
        elif isinstance(query, SqlSubQuery):
            self.add(query.stat)
        elif isinstance(query, SqlAlias) and isinstance(query._value, SqlSubQuery):
            self.add(query._value.stat)
            self.add(['AS', query._name])
        else:
            raise TypeError(repr(query))
        return self


class SqlDeleteStat(SqlStat[T], SqlWhereStat, SqlLimitStat, SqlReturnStat, Generic[T]):
    pass



import abc
import functools
import sqlite3
import sys
from pathlib import Path
from typing import Optional, Literal

import polars as pl

from neuralib.argp import AbstractParser, argument
from .connection import Connection

__all__ = ['Database', 'transaction']


def transaction():
    """
    A decorator that decorate a function as a transaction session
    that open a connection and set to the global variable.

    >>> @transaction
    ... def example(self):
    ...     self.do_something()

    equalivent to

    >>> def example(self):
    ...     with self.open_connection():
    ...         self.do_something()

    """

    def _decorator(f):
        @functools.wraps(f)
        def _transaction(self: Database, *args, **kwargs):
            with self.open_connection():
                return f(self, *args, **kwargs)

        return _transaction

    return _decorator


class Database(metaclass=abc.ABCMeta):
    """
    A class for the common usage of using sqlite.
    """

    sqlp_debug_mode = False

    @property
    @abc.abstractmethod
    def database_file(self) -> Optional[Path]:
        """sqlite database filepath"""
        pass

    @property
    @abc.abstractmethod
    def database_tables(self) -> list[type]:
        """supporting tables"""
        pass

    def open_connection(self) -> Connection:
        """
        open a connection to the database.
        """
        if (database_file := self.database_file) is None:
            database_file = ':memory:'
        else:
            database_file.parent.mkdir(exist_ok=True)

        ret = Connection(database_file, debug=self.sqlp_debug_mode)
        cls = type(self)
        if getattr(cls, '__first_connect_init', True):
            from .stat_start import create_table
            with ret:
                for table in self.database_tables:
                    create_table(table).submit()

            setattr(cls, '__first_connect_init', False)

        return ret


class CliDatabase(Database, AbstractParser):
    """
    Wrap Database to support commandline interface.
    """

    sqlp_debug_mode: bool = argument('--debug')

    DB_FILE: str = argument('-d', '--database', metavar='FILE')
    DB_STAT: list[str] = argument(metavar='STATS', nargs='*', action='extend')
    list_table: str = argument('--table', metavar='NAME', default=None, const='', nargs='?')
    from_file: bool = argument('-f', '--file')
    action: Literal['import', 'export'] = argument('--action', default=None)
    pretty: bool = argument('-p', '--pretty')

    USAGE = """\
%(prog)s -d FILE --table [NAME]
%(prog)s -d FILE STAT ...
%(prog)s -d FILE --file SCRIPT
%(prog)s -d FILE --action=(import|export) --table NAME FILE
"""

    def __init__(self, database: Optional[Database] = None):
        Database.__init__(self)
        self.database = database

    @property
    def database_file(self) -> Optional[Path]:
        if self.database is not None:
            return self.database.database_file
        return Path(self.DB_FILE)

    @property
    def database_tables(self) -> list[type]:
        if self.database is not None:
            return self.database.database_tables
        return []

    def open_connection(self) -> Connection:
        if self.database is not None:
            self.database.sqlp_debug_mode = self.sqlp_debug_mode
            return self.database.open_connection()
        return super().open_connection()

    def run(self):
        if self.action is not None:
            return self.run_action()

        if self.list_table is not None:
            return self.run_list_table()

        if self.from_file:
            return self.run_script()

        return self.run_statement()

    def run_list_table(self):
        with self.open_connection() as connection:
            if len(self.list_table) == 0:
                print(connection.list_table())
            else:
                print(connection.table_schema(self.list_table))

    def run_action(self):
        if self.action == 'export':
            if self.list_table is None or len(self.list_table) == 0:
                raise ValueError('missing --table')

            if len(self.DB_STAT) == 0:
                out = None
            elif len(self.DB_STAT) == 1:
                out = self.DB_STAT[0]
            else:
                raise RuntimeError(f'too many arguments : {self.DB_STAT[1:]}')

            with self.open_connection() as connection:
                ret = connection.export_dataframe(self.list_table).write_csv(out)
            if out is None:
                print(ret)

        elif self.action == 'import':
            if self.list_table is None or len(self.list_table) == 0:
                raise ValueError('missing --table')

            if len(self.DB_STAT) == 0:
                data = pl.read_csv(sys.stdin)
            elif len(self.DB_STAT) == 1:
                data = pl.read_csv(self.DB_STAT[0])
            else:
                raise RuntimeError(f'too many arguments : {self.DB_STAT[1:]}')

            with self.open_connection() as connection:
                connection.import_dataframe(self.list_table, data)

        else:
            raise ValueError(f'unknown action={self.action}')

    def run_script(self):
        if len(self.DB_STAT) > 1:
            raise RuntimeError(f'too many arguments : {self.DB_STAT[1:]}')

        with self.open_connection() as connection:
            stat = []

            with Path(self.DB_STAT[0]).open() as file:
                for line in file:
                    stat.append(line)

                    if ';' in line:
                        result = connection.execute(''.join(stat))
                        stat = []
                        self._print_result(connection, result)

            if len(stat):
                result = connection.execute(''.join(stat))
                self._print_result(connection, result)

    def run_statement(self):
        with self.open_connection() as connection:
            result = connection.execute(' '.join(self.DB_STAT))
            self._print_result(connection, result)

    def _print_result(self, connection, cursor: sqlite3.Cursor):
        if self.pretty:
            from neuralib.sqlp.stat import Cursor
            print(Cursor(connection, cursor).fetch_polars())
        else:
            header = cursor.description
            if header is not None:
                print('--', tuple([it[0] for it in header]))
            for data in cursor:
                print(data)

if __name__ == '__main__':
    CliDatabase().main()



from __future__ import annotations

from io import BufferedIOBase, BufferedReader
from pathlib import Path
from typing import BinaryIO, Union, TypeVar, Any

import numpy as np
import pandas as pd
import polars as pl
from matplotlib.axes import Axes
from numpy.typing import NDArray

__all__ = [
    'ArrayLike',
    'ArrayLikeStr',
    'NDArrayInt',
    'NDArrayFloat',
    'NDArrayBool',
    'AxesArrayLike',
    #
    'PathLike',
    #
    'Series',
    'DataFrame',
    #
    'is_iterable'
]

T = TypeVar('T')
ArrayLike = Union[np.ndarray, list[T], tuple[T, ...]]

ArrayLikeStr = Union[NDArray[np.str_], list[str], tuple[str, ...]]

NDArrayInt = NDArray[np.int_]
NDArrayFloat = NDArray[np.float_]
NDArrayBool = NDArray[np.bool_]

PathLike = Union[str, Path, bytes, BinaryIO, BufferedIOBase, BufferedReader]

#
Series = Union[pd.Series, pl.Series]
DataFrame = Union[pd.DataFrame, pl.DataFrame]

#
AxesArrayLike = Union[np.ndarray, list[Axes]]


# ============ #

def is_iterable(val: Any) -> bool:
    """check value is iterable"""
    from collections.abc import Iterable
    return isinstance(val, Iterable) and not isinstance(val, str)



"""
SQL functions related to datetime.

https://www.sqlite.org/lang_datefunc.html
"""
from typing_extensions import LiteralString

from . import expr
from .func_dec import as_func_expr

__all__ = [
    'date', 'time', 'datetime', 'julianday', 'unixepoch', 'strftime', 'timediff'
]


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def date(t, *m) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_datefunc.html

    :param t: time value
    :param m: modifier
    :return:
    """
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def time(t, *m) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_datefunc.html

    :param t: time value
    :param m: modifier
    :return:
    """
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def datetime(t, *m) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_datefunc.html

    :param t: time value
    :param m: modifier
    :return:
    """
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def julianday(t, *m) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_datefunc.html

    :param t: time value
    :param m: modifier
    :return:
    """
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def unixepoch(t, *m) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_datefunc.html

    :param t: time value
    :param m: modifier
    :return:
    """
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
def strftime(fmt: LiteralString, t, *m) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_datefunc.html

    :param fmt: format
    :param t: time value
    :param m: modifier
    :return:
    """
    return expr.SqlFuncOper('STRFTIME', expr.SqlLiteral(repr(fmt)), t, *m)


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def timediff(a, b) -> expr.SqlExpr:
    """

    https://www.sqlite.org/lang_datefunc.html

    :param a: time value
    :param b: time value
    :return:
    """
    pass



"""
SQL functions.

covers
* https://www.sqlite.org/lang_corefunc.html
* https://www.sqlite.org/lang_aggfunc.html
* https://www.sqlite.org/lang_mathfunc.html
"""
from __future__ import annotations

from typing import overload

from typing_extensions import LiteralString

from . import expr
from .func_dec import as_func_expr

__all__ = [
    'abs', 'changes', 'char', 'coalesce', 'format',
    'hex', 'ifnull', 'iff', 'instr', 'last_insert_rowid', 'length', 'likelihood', 'likely', 'lower', 'ltrim',
    'max', 'min', 'nullif', 'octet_length', 'printf', 'quote', 'random', 'randomblob', 'replace', 'round', 'rtrim',
    'sign', 'sqlite_source_id', 'sqlite_version', 'substr', 'total_changes', 'trim', 'typeof',
    'unhex', 'unicode', 'unlikely', 'upper', 'zeroblob',
    'avg', 'count', 'group_concat', 'sum',
    'acos', 'acosh', 'asin',
    'asinh', 'atan', 'atan2', 'atanh', 'ceil', 'cos', 'cosh', 'degrees', 'exp', 'floor', 'ln', 'log', 'log2', 'log10',
    'mod', 'pi', 'pow', 'radians', 'sin', 'sinh', 'sqrt', 'tan', 'tanh', 'trunc',
]


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def abs(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#abs"""
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def changes(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#changes"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def char(*x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#char"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def coalesce(*x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#coalesce"""
    pass


# https://www.sqlite.org/lang_corefunc.html#concat by func_stat
# https://www.sqlite.org/lang_corefunc.html#glob by func_stat

# noinspection PyShadowingBuiltins,PyUnusedLocal
def format(fmt: LiteralString, *x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#format"""
    return expr.SqlFuncOper('FORMAT', expr.SqlLiteral(repr(fmt)), *x)


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def hex(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#hex"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def ifnull(x, y) -> expr.SqlExpr:
    """
    ``x?:y``

    https://www.sqlite.org/lang_corefunc.html#ifnull

    """
    pass


# noinspection PyUnusedLocal
@as_func_expr
def iff(x, y, z) -> expr.SqlExpr:
    """
    ``x?y:z``

    https://www.sqlite.org/lang_corefunc.html#iif
    """
    pass


# noinspection PyUnusedLocal
@as_func_expr
def instr(x, y) -> expr.SqlExpr:
    """
    ``x.find(y)``

    https://www.sqlite.org/lang_corefunc.html#instr
    """
    pass


# noinspection PyUnusedLocal
@as_func_expr
def last_insert_rowid() -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#last_insert_rowid"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def length(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#length"""
    pass


# https://www.sqlite.org/lang_corefunc.html#like by func_stat

# noinspection PyUnusedLocal
@as_func_expr
def likelihood(x, y) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#likelihood"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def likely(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#likely"""
    pass


# TODO https://www.sqlite.org/lang_corefunc.html#load_extension

# noinspection PyUnusedLocal
@as_func_expr
def lower(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#lower"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def ltrim(x, y=None) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#ltrim"""
    pass


# noinspection PyShadowingBuiltins
@overload
def max(x) -> expr.SqlAggregateFunc:
    pass


# noinspection PyShadowingBuiltins
@overload
def max(x, *other) -> expr.SqlExpr:
    pass


# noinspection PyShadowingBuiltins
def max(x, *other):
    """
    https://www.sqlite.org/lang_corefunc.html#max
    https://www.sqlite.org/lang_aggfunc.html#max_agg
    """
    if len(other) == 0:
        return expr.SqlAggregateFunc('MAX', x)
    else:
        return expr.SqlFuncOper('MAX', x, *other)


# noinspection PyShadowingBuiltins
@overload
def min(x) -> expr.SqlAggregateFunc:
    pass


# noinspection PyShadowingBuiltins
@overload
def min(x, *other) -> expr.SqlExpr:
    pass


# noinspection PyShadowingBuiltins
def min(x, *other):
    """
    https://www.sqlite.org/lang_corefunc.html#min
    https://www.sqlite.org/lang_aggfunc.html#min_agg
    """
    if len(other) == 0:
        return expr.SqlAggregateFunc('MIN', x)
    else:
        return expr.SqlFuncOper('MIN', x, *other)


# noinspection PyUnusedLocal
@as_func_expr
def nullif(x, y) -> expr.SqlExpr:
    """
    ``x if x != y else None``

    https://www.sqlite.org/lang_corefunc.html#nullif
    """
    pass


# noinspection PyUnusedLocal
@as_func_expr
def octet_length(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#octet_length"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def printf(fmt: LiteralString, *x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#printf"""
    return expr.SqlFuncOper('PRINTF', expr.SqlLiteral(repr(fmt)), *x)


# noinspection PyUnusedLocal
@as_func_expr
def quote(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#quote"""
    pass


@as_func_expr
def random() -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#random"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def randomblob(n: int) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#randomblob"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def replace(x, y, z) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#replace"""
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def round(x, y=None) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#round"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def rtrim(x, y=None) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#rtrim"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def sign(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#sign"""
    pass


# TODO https://www.sqlite.org/lang_corefunc.html#soundex
# TODO https://www.sqlite.org/lang_corefunc.html#sqlite_compileoption_get
# TODO https://www.sqlite.org/lang_corefunc.html#sqlite_compileoption_used
# TODO https://www.sqlite.org/lang_corefunc.html#sqlite_offset

# noinspection PyUnusedLocal
@as_func_expr
def sqlite_source_id() -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#sqlite_source_id"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def sqlite_version() -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#sqlite_version"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def substr(x, y, z=None) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#substr"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def total_changes() -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#total_changes"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def trim(x, y=None) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#trim"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def typeof(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#typeof"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def unhex(x, y=None) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#unhex"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def unicode(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#unicode"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def unlikely(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#unlikely"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def upper(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#upper"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def zeroblob(n: int) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_corefunc.html#zeroblob"""
    pass


# noinspection PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def avg(x) -> expr.SqlAggregateFunc:
    """https://www.sqlite.org/lang_aggfunc.html#avg"""
    pass


def count(x=None) -> expr.SqlAggregateFunc:
    """https://www.sqlite.org/lang_aggfunc.html#count"""
    if x is None:
        return expr.SqlAggregateFunc('COUNT', expr.SqlLiteral('*'))
    else:
        return expr.SqlAggregateFunc('COUNT', x)


# noinspection PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def group_concat(x, y=None) -> expr.SqlAggregateFunc:
    """https://www.sqlite.org/lang_aggfunc.html#group_concat"""
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def sum(x) -> expr.SqlAggregateFunc:
    """https://www.sqlite.org/lang_aggfunc.html#sumunc"""
    pass


# noinspection PyUnusedLocal
@as_func_expr(func=expr.SqlAggregateFunc)
def total(x) -> expr.SqlAggregateFunc:
    """https://www.sqlite.org/lang_aggfunc.html#sumunc"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def acos(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#acos"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def acosh(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#acosh"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def asin(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#asin"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def asinh(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#asinh"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def atan(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#atan"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def atan2(y, x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#atan2"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def atanh(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#atanh"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def ceil(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#ceil"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def cos(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#cos"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def cosh(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#cosh"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def degrees(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#degrees"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def exp(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#exp"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def floor(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#floor"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def ln(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#ln"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def log2(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#log2"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def log10(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#log"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def log(b, x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#log"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def mod(x, y) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#mod"""
    pass


@as_func_expr
def pi() -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#pi"""
    pass


# noinspection PyShadowingBuiltins,PyUnusedLocal
@as_func_expr
def pow(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#pow"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def radians(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#radians"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def sin(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#sin"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def sinh(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#sinh"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def sqrt(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#sqrt"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def tan(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#tan"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def tanh(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#tanh"""
    pass


# noinspection PyUnusedLocal
@as_func_expr
def trunc(x) -> expr.SqlExpr:
    """https://www.sqlite.org/lang_mathfunc.html#trunc"""
    pass




from __future__ import annotations

from pathlib import Path
from typing import Literal, Callable

import numpy as np
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from scipy.ndimage import gaussian_filter1d

from neuralib.plot.colormap import insert_colorbar
from neuralib.util.util_type import ArrayLike, DataFrame, PathLike
from neuralib.util.util_verbose import fprint

__all__ = [
    'plot_2d_dots',
    'plot_regression_cc',
    'plot_joint_scatter_histogram',
    'plot_histogram_cutoff',
    'plot_half_violin_box_dot'
]


# ========= #
# Dots plot #
# ========= #

def plot_2d_dots(ax: Axes,
                 x: ArrayLike,
                 y: ArrayLike,
                 size: ArrayLike, *,
                 with_color: bool = False,
                 with_legends: bool = True,
                 size_type: Literal['area', 'radius'] = 'radius',
                 size_factor: float = 3000,
                 **kwargs):
    """
    Plot values as 2D dots

    :param ax: ``Axes``
    :param x: string values (X,)
    :param y: string values (Y,)
    :param size: size of the dots (Y, X)
    :param with_color: `size` domain as colormap
    :param with_legends: show the value scaling as legend
    :param size_type: whether the value corresponding to circle radius or surface area.
    :param size_factor: scaling factor for visualization
    :param kwargs: passed to ``ax.scatter()``
    :return:
    """
    size = np.array([size]).flatten()
    x, y = np.meshgrid(x, y, indexing='ij')
    #
    if size_type == 'radius':
        func = lambda it: it ** 2 * size_factor
    elif size_type == 'area':
        func = lambda it: it * size_factor
    else:
        raise ValueError('')

    s = func(size)
    #
    if with_color:
        im = ax.scatter(x.ravel(), y.ravel(), s=s, c=size, cmap='viridis', clip_on=False,
                        vmin=0, vmax=1, **kwargs)
        insert_colorbar(ax, im)
    else:
        ax.scatter(x.ravel(), y.ravel(), s=s, c='k', clip_on=False, **kwargs)

    if with_legends:
        ax_size_legend(ax, size, func)


def ax_size_legend(ax: Axes,
                   value: list[int] | np.ndarray,
                   f: Callable[[float], float] = None):
    """
    add the label and legend of the size in scatter plot

    :param ax: ``Axes``
    :param value: values reflect to size
    :param f: amplified callable
    :return:
    """
    if f is None:
        f = lambda it: (it ** 2) * 3000

    vsize = np.linspace(0, np.max(value), num=5)

    for s in vsize:
        ax.scatter([], [], s=f(s), c='k', label=str(s))

    h, l = plt.gca().get_legend_handles_labels()
    plt.legend(h[1:], l[1:], labelspacing=1.2, title="value", borderpad=1,
               frameon=True, framealpha=0.6, edgecolor="k", facecolor="w")


# ========== #
# Regression #
# ========== #


def plot_regression_cc(ax: Axes,
                       x: np.ndarray,
                       y: np.ndarray,
                       order: int = 1,
                       show_cc: bool = True,
                       bins: int = 10,
                       bin_func: Literal['median', 'mean'] = 'median',
                       **kwargs):
    """
    Regression to see the relationship between x and y

    :param ax: ``Axes``
    :param x: numerical array x
    :param y: numerical array x
    :param order: order of the polynomial to fit when calculating the residuals.
    :param bins: number of bins
    :param show_cc: show correlation coefficient
    :param bin_func: Literal['median', 'mean']. default is median
    :param kwargs: passed to ``ax.set``
    :return:
    """
    import seaborn as sns

    sns.regplot(x=x, y=y, ax=ax, order=order,
                scatter_kws={
                    'alpha': 0.5,
                    'color': 'grey',
                    's': 8,
                    'edgecolors': 'none'
                },
                line_kws={
                    'color': 'black'
                })
    _hist_line_plot(ax, x, y, bins, bin_func=bin_func)

    ax.set(**kwargs)
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')

    if show_cc:
        cc = calculate_cc(x, y)
        ax.set_title(f'r = {round(cc, 4)}', fontstyle='italic')


def _hist_line_plot(ax,
                    x: np.ndarray,
                    y: np.ndarray,
                    bins: int = 20,
                    bin_func: Literal['median', 'mean'] = 'median'):
    """
    average or pickup the median of the y value in certain bins
    :param ax: ``Axes``
    :param x: (N, )
    :param y: (N, )
    :param bins:
    :param bin_func
    :return:
    """

    if bin_func == 'mean':
        a, edg = np.histogram(x, weights=y, bins=bins)
        n = np.histogram(x, bins)[0]  # (B, )

        weights = np.divide(a, n, out=np.zeros_like(a, dtype=float), where=n != 0)  # avoid true divide
        weights = gaussian_filter1d(weights, 1)

        x = edg[:-1] + np.median(np.diff(edg)) / 2  # alignment
        ax.plot(x, weights, 'r--', alpha=0.5)

    elif bin_func == 'median':
        counts, edg = np.histogram(x, weights=y, bins=bins)

        medians = []
        for i in range(len(counts)):
            bin_data = y[(x >= edg[i]) & (x < edg[i + 1])]
            median_val = np.nanmedian(bin_data)
            medians.append(median_val)

        x = edg[:-1] + np.median(np.diff(edg)) / 2  # alignment
        ax.plot(x, medians, 'r--', alpha=0.5)

    else:
        raise ValueError(f'{bin_func}')


def calculate_cc(x: np.ndarray, y: np.ndarray) -> float:
    from scipy.stats import pearsonr
    try:
        cc = pearsonr(x, y)[0]
    except ValueError as e:
        print('input shape ERROR:', x.shape, y.shape)
        raise e

    return cc


# ====== #
# Others #
# ====== #


def plot_histogram_cutoff(ax: Axes,
                          value: ArrayLike,
                          cutoff: float,
                          mask: np.ndarray | None = None,
                          **kwargs) -> None:
    """
    Plot the histogram with a cutoff value

    :param ax: ``Axes``
    :param value: (N,) 1d array
    :param cutoff: cutoff (threshold) value for the certain value, >= represents pass
    :param mask: (N,) mask for value. i.e., cell selection
    :param kwargs: passed to ``ax.set``
    """
    if value.ndim != 1:
        raise ValueError('value must be 1d array')

    if mask is not None:
        value = value[mask]

    if cutoff > np.max(value) or cutoff < np.min(value):
        fprint(f'{cutoff} should be within {np.min(value)} and {np.max(value)}', vtype='warning')

    # ax.hist(value, 50, density=True, color='grey')
    sns.histplot(value, bins=30, kde=True, ax=ax, color='grey', stat='percent', element='step')
    ax.axvline(cutoff, color='r', linestyle='--', zorder=1)
    ax.set(**kwargs)
    ax.set_aspect(1.0 / ax.get_data_ratio(), adjustable='box')


def plot_joint_scatter_histogram(x: np.ndarray,
                                 y: np.ndarray,
                                 show_cc: bool = True,
                                 output: Path | None = None,
                                 **kwargs):
    """
    plot the linear correlation scatter and histogram between two variables

    :param x: numerical array x
    :param y: numerical array y
    :param show_cc: if show correlation coefficient
    :param output: fig save output
    :param kwargs: pass through ``ax.set()``
    :return:
    """

    sns.set(style='white', font_scale=1.2)

    g = sns.JointGrid(x=x, y=y, height=5)
    g = g.plot_joint(sns.regplot, color='grey')
    g = g.plot_marginals(sns.histplot, kde=False, bins=12, color='grey')

    ax = g.ax_joint
    ax.set(**kwargs)

    if show_cc:
        cc = calculate_cc(x, y)
        ax.text(0.5, 0.8, f'r = {round(cc, 4)}', fontstyle='italic',
                horizontalalignment='center',
                verticalalignment='center',
                transform=ax.transAxes)

    if output is not None:
        plt.savefig(output)
    else:
        plt.show()


def plot_half_violin_box_dot(ax: Axes,
                             data: DataFrame | dict | list[np.ndarray],
                             x: str | None = None,
                             y: str | None = None,
                             hue: str | None = None,
                             output: PathLike | None = None,
                             **kwargs) -> None:
    """
    Plot the data with half violin together with boxes and dots

    :param ax: ``Axes``
    :param data: Dataset for plotting
    :param x: names of variables in data or vector data: ``x``
    :param y: names of variables in data or vector data: ``y``
    :param hue: names of variables in data or vector data: ``hue``
    :param output: fig save output
    :param kwargs: pass through ``sns.violinplot``, ``sns.boxplot`` and ``sns.stripplot``
    :return:
    """
    kws = dict(
        ax=ax,
        x=x,
        y=y,
        hue=hue,
        data=data,
        palette='Set2',
        **kwargs
    )

    sns.violinplot(dodge=False, density_norm="width", inner=None, **kws)

    xlim = ax.get_xlim()
    ylim = ax.get_ylim()
    for violin in ax.collections:
        bbox = violin.get_paths()[0].get_extents()
        x0, y0, width, height = bbox.bounds
        violin.set_clip_path(plt.Rectangle((x0, y0), width / 2, height, transform=ax.transData))

    sns.boxplot(saturation=1, showfliers=False, width=0.3, boxprops={'zorder': 3, 'facecolor': 'none'}, **kws)

    old_len_collections = len(ax.collections)
    sns.stripplot(dodge=False, alpha=0.7, size=3, **kws)

    for dots in ax.collections[old_len_collections:]:
        dots.set_offsets(dots.get_offsets() + np.array([0.12, 0]))
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    if output is not None:
        plt.savefig(output)



"""
Wrapper for iblatlas
======================

Provide basic plotting usage

See detail in ``iblatlas.plots.plot_scalar_on_slice``

.. seealso:: `<https://int-brain-lab.github.io/iblenv/notebooks_external/atlas_plotting_scalar_on_slice.html>`_


.. toctree::
    :maxdepth: 1
    :caption: examples

    ../notebooks/example_ibl_plot

"""

from .plot import *
